/*
 *  Images.cpp
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "Images.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <FreeImage.h>

void IMAGERaw::SetSize( const int in_iX, const int in_iY, bool in_bHasAlpha )
{
    m_bHasAlpha = in_bHasAlpha;
    
    m_iXSize = in_iX;
    m_iYSize = in_iY;
    
    m_aiXIndx.resize( in_iY );
    if ( m_bHasAlpha == TRUE ) {
        m_iStep = 4;
    } else {
        m_iStep = 3;
    }
    m_afImage.resize( in_iX * in_iY * m_iStep );
    
    for (int i = 0; i < in_iY; i++)
		m_aiXIndx[i] = (in_iY - i - 1) * in_iX * m_iStep;
    
}


IMAGERaw::IMAGERaw()
{
    m_bHasAlpha = FALSE;
    SetSize(0,0, FALSE);
    
}

IMAGERaw::IMAGERaw(const int in_iX, const int in_iY, const bool in_bAlpha)
{
    
    SetSize(in_iX, in_iY, in_bAlpha);
}

IMAGERaw::~IMAGERaw()
{
    SetSize(0,0);
}


void IMAGERaw::Set( const UTILSColor &in_pix )
{
    const int iSize = Width() * Height();
    for ( int i = 0; i < iSize; i++ ) {
        const int iOff = i * m_iStep;
        m_afImage[RED + iOff]   = (float) in_pix[0];
        m_afImage[GREEN + iOff] = (float) in_pix[1];
        m_afImage[BLUE + iOff]  = (float) in_pix[2];
    }
}


IMAGERaw &IMAGERaw::operator=( const IMAGERaw &in_image ) 
{
    SetSize( in_image.Width(), in_image.Height(), in_image.HasAlpha() );
    m_afImage = in_image.m_afImage;
    
    return *this;
}

bool IMAGERaw::Read( const char * in_strFName )
{
    FreeImage_Initialise();
    
    std::string str( in_strFName );
    const int iExt = str.find_last_of( "." );
    std::string strExt = str.erase( 0, iExt + 1 );
    char sExt[10];
    strcpy( sExt, strExt.c_str() );
    
    for ( unsigned int i = 0; i < strlen( sExt ); i++ ) 
        sExt[i] = toupper( sExt[i] );
    
    if ( !strcmp( sExt, "TIF" ) )
        strcpy( sExt, "TIFF" );
    
    FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFormat( sExt );
    
    if ( fif == FIF_UNKNOWN ) {
        ASSERT(FALSE);
        return FALSE;
    }
    
    FIBITMAP *bm = FreeImage_Load( fif, in_strFName );
    
    if ( !bm )
        return FALSE;
    
    FREE_IMAGE_COLOR_TYPE imColType = FreeImage_GetColorType( bm );
    SetSize( FreeImage_GetWidth(bm), FreeImage_GetHeight(bm), imColType == FIC_RGBALPHA ? TRUE : FALSE );
    
    RGBQUAD rgb;
    for ( int iX = 0; iX < Width(); iX++ ) {
        for ( int iY = 0; iY < Height(); iY++ ) {
            FreeImage_GetPixelColor( bm, iX, iY, &rgb );
            Set( iX, Height() -1 -iY, UTILSColor( (double) rgb.rgbRed / 255.0, (double) rgb.rgbGreen / 255.0, (double) rgb.rgbBlue / 255.0 ) );
        }
    }
    
    FreeImage_Unload( bm );
    
    return TRUE;
}


bool IMAGERaw::Write( const char * in_strFName, const bool in_bAlpha ) const
{
    FreeImage_Initialise();
    
    std::string str( in_strFName );
    const int iExt = str.find_last_of( "." );
    std::string strExt = str.erase( 0, iExt + 1 );
    char sExt[10];
    strcpy( sExt, strExt.c_str() );
    
    for ( unsigned int i = 0; i < strlen( sExt ); i++ ) 
        sExt[i] = toupper( sExt[i] );
    
    if ( !strcmp( sExt, "TIF" ) )
        strcpy( sExt, "TIFF" );
    
    FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFormat( sExt );
    
    if ( fif == FIF_UNKNOWN ) {
        ASSERT(FALSE);
        return FALSE;
    }
    
    FIBITMAP *bm = FreeImage_AllocateT( FIT_BITMAP, Width(), Height(), 32 );
    
    RGBQUAD rgb;
    for ( int iX = 0; iX < Width(); iX++ ) {
        for ( int iY = 0; iY < Height(); iY++ ) {
            const UTILSColor col = (*this)( iX, iY );
            rgb.rgbRed = (unsigned int) (col[0] * 255.0);
            rgb.rgbGreen = (unsigned int) (col[1] * 255.0);
            rgb.rgbBlue = (unsigned int) (col[2] * 255.0);
            FreeImage_SetPixelColor( bm, iX, Height() -1 -iY, &rgb );
        }
    }
    
    FreeImage_Save( fif, bm, in_strFName );
    
    FreeImage_Unload( bm );
    
    return TRUE;
}

void IMAGERaw1D::SetSize( const int in_iX, const int in_iY )
{
    m_iXSize = in_iX;
    m_iYSize = in_iY;
    
    m_aiXIndx.resize( in_iY );
    m_afImage.resize( in_iX * in_iY );
    
    for (int i = 0; i < in_iY; i++)
		m_aiXIndx[i] = (in_iY - i - 1) * in_iX;
    
}

IMAGERaw1D::IMAGERaw1D()
{
    SetSize(0,0);
    
}

IMAGERaw1D::IMAGERaw1D(const int in_iX, const int in_iY)
{
    SetSize(in_iX, in_iY);
}

IMAGERaw1D::~IMAGERaw1D()
{
    SetSize(0,0);
}


void IMAGERaw1D::Set( const float &in_pix )
{
    const int iSize = Width() * Height();
    for ( int i = 0; i < iSize; i++ ) {
        m_afImage[i]   = in_pix;
    }
}


bool IMAGERaw1D::Read( const char * in_strFName )
{
    FreeImage_Initialise();
    
    std::string str( in_strFName );
    const int iExt = str.find_last_of( "." );
    std::string strExt = str.erase( 0, iExt + 1 );
    char sExt[10];
    strcpy( sExt, strExt.c_str() );
    
    for ( unsigned int i = 0; i < strlen( sExt ); i++ ) 
        sExt[i] = toupper( sExt[i] );
    
    if ( !strcmp( sExt, "TIF" ) )
        strcpy( sExt, "TIFF" );
    
    FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFormat( sExt );
    
    if ( fif == FIF_UNKNOWN ) {
        ASSERT(FALSE);
        return FALSE;
    }
    
    FIBITMAP *bm = FreeImage_Load( fif, in_strFName );
    if ( bm == NULL )
        return FALSE;
    
    SetSize( FreeImage_GetWidth(bm), FreeImage_GetHeight(bm) );
    
    RGBQUAD rgb;
    for ( int iX = 0; iX < Width(); iX++ ) {
        for ( int iY = 0; iY < Height(); iY++ ) {
            FreeImage_GetPixelColor( bm, iX, iY, &rgb );
            const float dAvg = ( (float) rgb.rgbRed + (float) rgb.rgbGreen + (float) rgb.rgbBlue ) / (255.0f * 3.0f);
            Set( iX, Height() -1 - iY, dAvg );
        }
    }
    
    FreeImage_Unload( bm );
    
    return TRUE;
}


bool IMAGERaw1D::Write( const char * in_strFName ) const
{
    FreeImage_Initialise();
    
    std::string str( in_strFName );
    const int iExt = str.find_last_of( "." );
    std::string strExt = str.erase( 0, iExt + 1 );
    char sExt[10];
    strcpy( sExt, strExt.c_str() );
    
    for ( unsigned int i = 0; i < strlen( sExt ); i++ ) 
        sExt[i] = toupper( sExt[i] );
    
    if ( !strcmp( sExt, "TIF" ) )
        strcpy( sExt, "TIFF" );
    
    FREE_IMAGE_FORMAT fif = FreeImage_GetFIFFromFormat( sExt );
    
    if ( fif == FIF_UNKNOWN ) {
        ASSERT(FALSE);
        return FALSE;
    }
    
    FIBITMAP *bm = FreeImage_AllocateT( FIT_BITMAP, Width(), Height(), 32 );
    
    RGBQUAD rgb;
    for ( int iX = 0; iX < Width(); iX++ ) {
        for ( int iY = 0; iY < Height(); iY++ ) {
            const double dCol = (*this)( iX, iY );
            rgb.rgbRed = (unsigned int) (dCol * 255.0);
            rgb.rgbGreen = (unsigned int) (dCol * 255.0);
            rgb.rgbBlue = (unsigned int) (dCol * 255.0);
            FreeImage_SetPixelColor( bm, iX, Height() -1 -iY, &rgb );
        }
    }
    
    FreeImage_Save( fif, bm, in_strFName );
    
    FreeImage_Unload( bm );
    
    return TRUE;
}


IMAGERaw1D &IMAGERaw1D::operator=( const IMAGERaw1D &in_image ) 
{
    SetSize( in_image.Width(), in_image.Height() );
    m_afImage = in_image.m_afImage;
    
    return *this;
}

void OGLImages_GetImage( IMAGERaw &out_oImage ) 
{
    int ai[4], iWidth, iHeight, iN;
    glGetIntegerv( GL_VIEWPORT, ai );
    
    iWidth = ai[2];
    iHeight = ai[3];
    iN = iWidth * iHeight * 3;
    
    out_oImage.SetSize( iWidth, iHeight, FALSE );
    
    unsigned char *buf = new unsigned char [iN];
    
    ::glReadPixels( 0, 0, iWidth, iHeight, GL_RGB, GL_FLOAT, (GLvoid *) out_oImage.Data() );
    /*
     float *af = out_oImage.Data();
     for ( int i = 0; i < iN; i++ ) {
     af[i] = (float) (buf[i] ) / 255.0f;
     }
     */
    
    delete [] buf;
}


