/*
 *  OGLRoutines.h
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */
#ifndef _OGL_ROUTINES_H
#define _OGL_ROUTINES_H

#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "SystemSpecific.h"
#include "UTILSColor.h"
#include "Geometry.h"

extern void CgErrorCallback();

/** \class OGLShader2 OGLShader.h OpenGL/OGLShader.h
 \ingroup Drawing
 \brief A wrapper around vertex and fragment shaders
 
 This class is a starting point for writing a shader in the FlTK viewer
 */
class OGLShader2 {
public:
	OGLShader2();
	~OGLShader2();

	bool LoadShaderFromFile(const char *pFileName);
	bool SetTechnique(char *pTech);
	bool StartPass();
	bool EndPass();

	bool SetOGLMatrix4x4BySemantic(char * pSemantic, const R4Matrix &pMat);
	bool SetTextureBySemantic(char * pSemantic, GLuint pID);
	bool SetFloat3BySemantic(char * pSemantic, float pVec[3]);
	bool SetFloatBySemantic(char * pSemantic, float pValue);
	bool SetBoolBySemantic(char * pSemantic, bool pValue);
	bool SetIntBySemantic(char * pSemantic, int pValue);

	bool Validate();
	const char * GetTechnique(int pIndex);
    
private:
	CGeffect effect;
	CGtechnique technique;
	CGcontext context;
	CGpass pass;
	bool verbose;
	int numPasses;
	int numTechniques;
	bool loaded;
	bool complete;
	bool firstPass;
	vector<CGtechnique> techniqueList;
	vector<const char *> namesList;
	vector< pair<GLuint, CGparameter> > m_aTexIds;
};

/** \defgroup Drawing2D Drawing 2D objects (squares, polygons, etc)
 \ingroup Drawing */
//@{
/// Sets the view matrix to be the identity, disables lighting
extern void OGLBegin2D( const int in_iWidth, const int in_iHeight );
///
extern void OGLDrawCircle( const R2Pt &in_pt, const double in_dRadius );

/// Pops the matrix, enables lighting
extern void OGLEnd2D( );
//@}


#endif