// Painter3D.cpp: implementation of the Painter3D class.
//
//////////////////////////////////////////////////////////////////////

#include <FL/gl.h>
#include "Painter3D.H"
#include "UserInterface.h"

/// Force enough redraws
int Painter3D::s_iCountRedraw = 101;

Painter3D::Painter3D(int X, int Y, int W, int H, const char *L) : 
Fl_Gl_Window(X, Y, W, H, L),
m_opShader(NULL),
m_bDirtyNorm(true),
m_bDirtyShift(true),
m_cScatterId(0),
m_iFrame(-1),
m_bPlayMovie(false), 
m_opUI(NULL),
m_dPercScatter(1.0),
m_imageScatterPrev(4,4), m_imageScatterNext(4,4), m_imageScatterCur(4,4),
m_strEnvMapName( "../Data/EnvironmentMaps/PastelEnvMap.tif" ), 
m_strImageName( ("../Data/3DPaintings/Street") ), m_strFileExtension( ".tif" ),
m_ptLastLight(0,0), 
m_ptNextLight(0,0),
m_ptLastCamera(0,0),
m_ptNextCamera(0,0),
m_dRandomWalk(0),
m_bInDrawRoutine(false)
{	
    mode( FL_RGB | FL_DOUBLE | FL_ALPHA );


    for ( int iT = 0; iT < NTypes3D(); iT++ ) {
        IMAGERaw &image = Image3D( GetType3D(iT) );
        image.SetSize( 64, 64 );
        const UTILSColor col1 = UTILSColor::WHITE;
        const UTILSColor col2 = UTILSColor::BLACK;

        for ( int iX = 0; iX < 64; iX++ ) {
            for ( int iY = 0; iY < 64; iY++ ) {
                image.Set( iX, iY, ( (iX / 8) % 2 || (iY / 8) % 2 == 0 ) ? col1 : col2 );
            }
        }
        SetOriginalImageSize( GetType3D(iT) ) = R2Pt_i(64, 64);
    }
    m_imageScatterPrev.Set( UTILSColor::WHITE );
    m_imageScatterNext.Set( UTILSColor::BLACK );

    for ( FORINT iT = 0; iT < NTypes1D(); iT++ ) {
        IMAGERaw1D &image = Image1D( GetType1D(iT) );
        image.SetSize( 64, 64 );

        for ( int iX = 0; iX < 64; iX++ ) {
            for ( int iY = 0; iY < 64; iY++ ) {
                image.Set( iX, iY, ( (iX / 8) % 2 || (iY / 8) % 2 == 0 ) ? 0 : 1.0 );
            }
        }
        SetOriginalImageSize( GetType1D(iT) ) = R2Pt_i(64, 64);
    }

    ReadEnvironmentMap( "../Data/EnvironmentMaps/PastelEnvMap.tif" );

    Redraw();
}

Painter3D::~Painter3D()
{
}


void Painter3D::StartRandomWalk() 
{ 
    m_dRandomWalk = 0.0; 
    m_ptLastCamera[0] = m_opUI->m_dCameraLR->value();
    m_ptLastCamera[1] = m_opUI->m_dCameraUD->value();
    m_ptLastLight[0] = m_opUI->m_dLightLR->value();
    m_ptLastLight[1] = m_opUI->m_dLightUD->value();
}

void Painter3D::CalcNextRandomFrame() 
{
    if ( m_dRandomWalk > 1.0 ) {
        m_dRandomWalk = 0.0;
        m_ptLastCamera = m_ptNextCamera;
        m_ptLastLight = m_ptNextLight;
        m_ptNextCamera[0] = 1.8 * ( rand() % 1024 ) / 1024.0 - 0.9;
        m_ptNextCamera[1] = 1.8 * ( rand() % 1024 ) / 1024.0 - 0.9;
        m_ptNextLight[0] = 1.8 * ( rand() % 1024 ) / 1024.0 - 0.9;
        m_ptNextLight[1] = 1.8 * ( rand() % 1024 ) / 1024.0 - 0.9;
    }
    
    const double dThetaCameraLast = atan2( m_ptLastCamera[1], m_ptLastCamera[0] );
    const double dThetaCameraNext = atan2( m_ptNextCamera[1], m_ptNextCamera[0] );
    const double dThetaLightLast = atan2( m_ptLastLight[1], m_ptLastLight[0] );
    const double dThetaLightNext = atan2( m_ptNextLight[1], m_ptNextLight[0] );
    
    const double dRadiusCameraLast = Length( m_ptLastCamera );
    const double dRadiusCameraNext = Length( m_ptNextCamera );
    const double dRadiusLightLast = Length( m_ptLastLight );
    const double dRadiusLightNext = Length( m_ptNextLight );
    
    double dThetaCamera = m_dRandomWalk * dThetaCameraNext + (1.0 - m_dRandomWalk) * dThetaCameraLast;
    double dThetaLight = m_dRandomWalk * dThetaLightNext + (1.0 - m_dRandomWalk) * dThetaLightLast;
    const double dRadiusCamera = m_dRandomWalk * dRadiusCameraNext + (1.0 - m_dRandomWalk) * dRadiusCameraLast;
    const double dRadiusLight = m_dRandomWalk * dRadiusLightNext + (1.0 - m_dRandomWalk) * dRadiusLightLast;
    
    if ( fabs( dThetaCameraLast - dThetaCameraNext ) > M_PI ) {
        if ( dThetaCameraLast > dThetaCameraNext )
            dThetaCamera = m_dRandomWalk * ( dThetaCameraNext + 2.0 * M_PI ) + (1.0 - m_dRandomWalk) * dThetaCameraLast;
        else
            dThetaCamera = m_dRandomWalk * dThetaCameraNext + (1.0 - m_dRandomWalk) * ( dThetaCameraLast + 2.0 * M_PI );
    }
    if ( fabs( dThetaLightLast - dThetaLightNext ) > M_PI ) {
        if ( dThetaLightLast > dThetaLightNext )
            dThetaLight = m_dRandomWalk * ( dThetaLightNext + 2.0 * M_PI ) + (1.0 - m_dRandomWalk) * dThetaLightLast;
        else
            dThetaLight = m_dRandomWalk * dThetaLightNext + (1.0 - m_dRandomWalk) * ( dThetaLightLast + 2.0 * M_PI );
    }
    const R2Pt ptCamera( dRadiusCamera * cos( dThetaCamera ), dRadiusCamera * sin( dThetaCamera ) );
    const R2Pt ptLight( dRadiusLight * cos( dThetaLight ), dRadiusLight * sin( dThetaLight ) );

    m_opUI->m_dCameraLR->value( ptCamera[0] );
    m_opUI->m_dCameraUD->value( ptCamera[1] );
    m_opUI->m_dLightLR->value( ptLight[0] );
    m_opUI->m_dLightUD->value( ptLight[1] );
    
    ChangedCamera();
}

/* Force enough redraws until screen is updated.
   s_iCountRedraw is decremented in ::draw(), incremented by Redraw() */
void Painter3D::Redraw_cb( void *in_data )
{
    Painter3D *opMe = static_cast<Painter3D *>( in_data );
    
    /// Put another callback on the stack in case the last one cleared
    if ( opMe->s_iCountRedraw > 0 ) {
        // We haven't finished drawing - put another callback on the stack just in case
        Fl::repeat_timeout( 0.00001, Painter3D::Redraw_cb, (void *) opMe );
        
        opMe->redraw();
    } else if ( opMe->m_bPlayMovie || opMe->m_opUI->m_bShowLayer->value() || opMe->m_opUI->m_bRandomWalk->value()  ) {
        // We successfully finished drawing
        // Need to redraw if either play movie or show layer is checked
        opMe->s_iCountRedraw = 1; // fltk workaround
        opMe->m_dPercScatter += 0.075; // update scatter
        opMe->m_dRandomWalk += 0.05; // update lights/camera
        Fl::repeat_timeout( 0.00001, Painter3D::Redraw_cb, (void *) opMe );
        
        if ( opMe->m_opUI->m_bRandomWalk->value() ) {
            opMe->CalcNextRandomFrame();
        }
        opMe->redraw();
    }
}

void Painter3D::draw()
{
    static bool s_bInInit = false;

    if ( s_bInInit == true )
        return;

    if ( m_bInDrawRoutine )
        return;
    
    m_bInDrawRoutine = true;
    if ( m_opShader == NULL ) {

        s_bInInit = true;
        ReadShader();

        glClearColor( 0.0f, 0.0f, 0.0f, 0.0 );

        ReadImages( "../Data/3DPaintings/Street" );
        //ReadImages( WINDataHome("3DPaintings/FigureNonSquare") );
        //ReadImages( WINDataHome("3DPaintings/Kia") );
        //ReadImages( WINDataHome("3DPaintings/Sunglasses") );
        //ReadImages( WINDataHome("3DPaintings/Painting") );
        s_bInInit = false;
    }

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    m_opShader->SetFloatBySemantic( "MaxHeight", m_opUI->m_dHeightScale->value() );
    m_opShader->SetFloatBySemantic( "StartShininess", m_opUI->m_dShininessStart->value() );
    m_opShader->SetFloatBySemantic( "MaxShininess", m_opUI->m_dShininessScale->value() );
    m_opShader->SetFloatBySemantic( "MaxDiffusion", m_opUI->m_dDiffusionScale->value() );
    m_opShader->SetFloatBySemantic( "MaxTransparency", m_opUI->m_dTransparencyScale->value() );

    //if ( m_opUI->m_bDoShadow->value() )
        //MakeShadowMaps();

    if ( m_opUI->m_bShowColor->value() ) {
        CreateHeightMap();
        DrawPaintColor();
    } else if ( m_opUI->m_bShowHeight->value() ) {
        CreateHeightMap();
        DrawHeightBW();
    } else if ( m_opUI->m_bShowHeightColor->value() ) {
        CreateHeightMap();
        DrawHeightColor();
    } else if ( m_opUI->m_bShowLayer->value() ) {
        CreateHeightMap();
        DrawPaintLayer();
    } else if ( m_opUI->m_bShowImage->value() ) {
        DrawImage();
    } else {
        ASSERT(false);
    }

    DrawCameraLightCircles();

    WriteMovieImage();
    s_iCountRedraw--;
    m_bInDrawRoutine = false;
}



int Painter3D::handle(int event) 
{

    static bool s_bCameraSelected = false;
    static bool s_bLightSelected = false;

    R2Pt ptCamera, ptLight;
    const R2Pt ptMouse( ( (double) Fl::event_x() / (double) w() ) * 2.0 - 1.0,  ( (double) Fl::event_y() / (double) h() ) * 2.0 - 1.0 );
    double dDistCam = 1e30, dDistLight = 1e30;
    switch(event) {
    case FL_PUSH:
        if ( m_opUI->m_bShowCamera->value() ) {
            ptCamera[0] = m_opUI->m_dCameraLR->value();    
            ptCamera[1] = m_opUI->m_dCameraUD->value();    
            dDistCam = Length( ptCamera, ptMouse );
        }
        if ( m_opUI->m_bShowLight->value() ) {
            ptLight[0] = m_opUI->m_dLightLR->value();    
            ptLight[1] = m_opUI->m_dLightUD->value();    
            dDistLight = Length( ptLight, ptMouse );
        }
        if ( dDistCam < dDistLight && dDistCam < 0.1 ) {
            s_bLightSelected = false;
            s_bCameraSelected = true;
        }
        if ( dDistLight < dDistCam && dDistLight < 0.1 ) {
            s_bLightSelected = true;
            s_bCameraSelected = false;
        }
            /*
        if ( m_opUI->m_bDrawing->value() ) {
            s_bLightSelected = false;
            s_bCameraSelected = false;
            s_bDrawing = true;
        }
             */
        return 1;

    case FL_MOVE : {
        //if ( m_opUI->m_bDrawing->value() )
        //    Redraw();
        return 1;
                   }

    case FL_DRAG: {
        if ( s_bCameraSelected ) {
            m_opUI->m_dCameraLR->value( ptMouse[0] );
            m_opUI->m_dCameraUD->value( ptMouse[1] );
            ChangedCamera();
            Redraw();
        }
        if ( s_bLightSelected ) {
            m_opUI->m_dLightLR->value( ptMouse[0] );
            m_opUI->m_dLightUD->value( ptMouse[1] );
            Redraw();
        }
        return 1;
        }
    case FL_RELEASE:   
        s_bLightSelected = false;
        s_bCameraSelected = false;
        s_bCameraSelected = false;
        return 1;
    case FL_FOCUS :
    case FL_UNFOCUS :
        // Return 1 if you want keyboard events, 0 otherwise
        return 1;
    case FL_LEAVE :
        break;
    case FL_KEYUP:
        return 1;

    default:
        // pass other events to the base class...
        return Fl_Gl_Window::handle(event);
    }
    return Fl_Gl_Window::handle(event);
}

void Painter3D::resize( int x, int y, int w, int h )
{
    Fl_Gl_Window::resize( x, y, w, h );

    Redraw();
}

Painter3D::ImageType Painter3D::GetType1D( const int in_i ) const
{
    static ImageType s_atype[4] = { SHININESS, TRANSPARENCY, DIFFUSION, HEIGHT };
    return s_atype[in_i];
}

Painter3D::ImageType Painter3D::GetType3D( const int in_i ) const
{
    static ImageType s_atype[5] = { DIFFUSE, SPECULAR, ENVIRONMENT_MAP, IRRADIANCE_MAP, PAPER };
    return s_atype[in_i];
}


GLuint Painter3D::TextureId( const ImageType &in_type ) const
{
    switch ( in_type ) {
    case DIFFUSE :
        return m_acTextureId[ 0 ];
    case SPECULAR :
        return m_acTextureId[ 1 ];
    case SHININESS :
        return m_acTextureId[ 2 ];
    case TRANSPARENCY :
        return m_acTextureId[ 3 ];
    case DIFFUSION :
        return m_acTextureId[ 4 ];
    case HEIGHT :
        return m_acTextureId[ 5 ];
    case PAPER :
        return m_acTextureId[ 6 ];
    case ENVIRONMENT_MAP :
        return m_acEnvMapId[ 0 ];
    case IRRADIANCE_MAP :
        return m_acEnvMapId[ 1 ];
    }
    ASSERT(false);
    return 0;
}

IMAGERaw1D &Painter3D::Image1D( const ImageType &in_type ) 
{
    switch ( in_type ) {
    case SHININESS :
        return m_imageShininess;
    case TRANSPARENCY :
        return m_imageTransparency;
    case DIFFUSION :
        return m_imageDiffusuion;
    case HEIGHT :
        return m_imageHeight;
    }
    ASSERT(false);
    return m_imageShininess;
}

const IMAGERaw1D &Painter3D::GetImage1D( const ImageType &in_type ) const
{
    switch ( in_type ) {
    case SHININESS :
        return m_imageShininess;
    case TRANSPARENCY :
        return m_imageTransparency;
    case DIFFUSION :
        return m_imageDiffusuion;
    case HEIGHT :
        return m_imageHeight;
    }
    ASSERT(false);
    return m_imageShininess;
}

IMAGERaw &Painter3D::Image3D( const ImageType &in_type ) 
{
    switch ( in_type ) {
    case DIFFUSE :
        return m_imageDiffuse;
    case SPECULAR :
        return m_imageSpecular;
    case ENVIRONMENT_MAP :
        return m_imageEnvMap;
    case IRRADIANCE_MAP :
        return m_imageIrrMap;
    case PAPER :
        return m_imagePaper;
    }
    ASSERT(false);
    return m_imageEnvMap;
}

const IMAGERaw &Painter3D::GetImage3D( const ImageType &in_type ) const
{
    switch ( in_type ) {
    case DIFFUSE :
        return m_imageDiffuse;
    case SPECULAR :
        return m_imageSpecular;
    case ENVIRONMENT_MAP :
        return m_imageEnvMap;
    case IRRADIANCE_MAP :
        return m_imageIrrMap;
    case PAPER :
        return m_imagePaper;
    }
    ASSERT(false);
    return m_imageEnvMap;
}

const char *Painter3D::ImageName( const ImageType &in_type ) const
{
    static char s_strs[9][15] = 
    { "diffuse", "specular", "shininess", "transparency", "diffusion", "height", "env", "irr", "paper" };

    static char s_str[15];
    switch ( in_type ) {
    case DIFFUSE :
        sprintf( s_str, "%s", s_strs[0] );
        return s_str;
    case SPECULAR :
        sprintf( s_str, "%s", s_strs[1] );
        return s_str;
    case SHININESS :
        sprintf( s_str, "%s", s_strs[2] );
        return s_str;
    case TRANSPARENCY :
        sprintf( s_str, "%s", s_strs[3] );
        return s_str;
    case DIFFUSION :
        sprintf( s_str, "%s", s_strs[4] );
        return s_str;
    case HEIGHT :
        sprintf( s_str, "%s", s_strs[5] );
        return s_str;
    case ENVIRONMENT_MAP :
        return s_strs[6];
    case IRRADIANCE_MAP :
        return s_strs[7];
    case PAPER :
        return s_strs[8];
    }
    ASSERT(false);
    return s_str;

}

const R2Pt_i &Painter3D::GetOriginalImageSize( const ImageType &in_type ) const
{
    switch ( in_type ) {
        case DIFFUSE :
            return m_aiptSaveImageSize[0];
        case SPECULAR :
            return m_aiptSaveImageSize[1];
        case SHININESS :
            return m_aiptSaveImageSize[2];
        case TRANSPARENCY :
            return m_aiptSaveImageSize[3];
        case DIFFUSION :
            return m_aiptSaveImageSize[4];
        case HEIGHT :
            return m_aiptSaveImageSize[5];
        case ENVIRONMENT_MAP :
            return m_aiptSaveImageSize[6];
        case IRRADIANCE_MAP :
            return m_aiptSaveImageSize[7];
        case PAPER :
            return m_aiptSaveImageSize[8];
    }
    ASSERT(false);
    
    return m_aiptSaveImageSize[0];
}

R2Pt_i &Painter3D::SetOriginalImageSize( const ImageType &in_type ) 
{
    if ( m_aiptSaveImageSize.size() != NImagesTotal() )
        m_aiptSaveImageSize.resize( NImagesTotal() );
    
    switch ( in_type ) {
        case DIFFUSE :
            return m_aiptSaveImageSize[0];
        case SPECULAR :
            return m_aiptSaveImageSize[1];
        case SHININESS :
            return m_aiptSaveImageSize[2];
        case TRANSPARENCY :
            return m_aiptSaveImageSize[3];
        case DIFFUSION :
            return m_aiptSaveImageSize[4];
        case HEIGHT :
            return m_aiptSaveImageSize[5];
        case ENVIRONMENT_MAP :
            return m_aiptSaveImageSize[6];
        case IRRADIANCE_MAP :
            return m_aiptSaveImageSize[7];
        case PAPER :
            return m_aiptSaveImageSize[8];
    }
    ASSERT(false);
    
    return m_aiptSaveImageSize[0];
}



double Painter3D::HeightScale() const 
{ 
    if (m_opUI) 
        return m_opUI->m_dHeightScale->value(); 
    return 0.0; 
}
