// generated by Fast Light User Interface Designer (fluid) version 1.0100

#ifndef Painter3D_h
#define Painter3D_h


#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Gl_Window.H>

#include "SystemSpecific.h"
#include "Geometry.h"
#include "UTILSColor.h"
#include "Images.h"
#include "OGLRoutines.h"

class UserInterface;
/** \brief Main viewing window

  Has support fo drawing meshes, points/dots, the IBar, and shaders
  */
class Painter3D : public Fl_Gl_Window {
public:
    typedef enum {
        DIFFUSE,
        SPECULAR,
        SHININESS,    // Specular coefficient
        HEIGHT,
        TRANSPARENCY, // 
        DIFFUSION,    // 
        PAPER,
        ENVIRONMENT_MAP,
        IRRADIANCE_MAP,
        SHADOW_MAP,
        ALL
    } ImageType;

protected:    

    /// The FlTK ui
    UserInterface *m_opUI;  
    
    class DisplayData {
    private:
        double m_dHeightScale, m_dTranspScale, m_dDiffuseScale, m_dShinyScale;
        double m_dShinyMin, m_dRefractMax, m_dRefractIndex, m_dRefractSpeed;
        double m_dCameraDist;
        R2Pt m_ptLight, m_ptCamera;
        bool m_bUseLight;
        UTILSColor m_colLight, m_colShadow;
    public:
        void PullFromUI( const UserInterface *in_opUI );
        void PushToUI( UserInterface *io_opUI ) const;

        void Write( ofstream &out ) const;
        void Read( ifstream &in);

        bool operator==( const DisplayData & ) const;
        DisplayData &operator=( const DisplayData & );
        DisplayData( const DisplayData & );
        DisplayData();
        ~DisplayData() {}
    };

            
    /// To fix FlTk's redraw problem
    static int s_iCountRedraw;

    GLuint m_cScatterId;
    vector< GLuint > m_acTextureId, m_acEnvMapId;  ///< OpenGL Id to refer to texture

    OGLShader2 *m_opShader;

    IMAGERaw m_imageDiffuse, 
             m_imageSpecular,
             m_imageEnvMap,
             m_imageIrrMap,
             m_imagePaper; 

    IMAGERaw1D m_imageShininess,
               m_imageHeight,
               m_imageTransparency,
               m_imageDiffusuion; // Height, transparency, and diffusion

    bool        m_bDirtyNorm, m_bDirtyShift;
    R2Pt_i          m_iptGridSize;
    R2Pt            m_ptTexStart, m_ptTexStep;
    vector<R3Vec_f> m_avecNorm;
    vector<R3Pt_f>  m_aptQuad;
    vector<R2Pt_f>  m_aptShiftUV;

    vector<R2Pt_i> m_aiptSaveImageSize;

    // Random walk
    R2Pt m_ptLastLight, m_ptNextLight;
    R2Pt m_ptLastCamera, m_ptNextCamera;
    double m_dRandomWalk;
    
    // Scatter texture
    double    m_dPercScatter;
    IMAGERaw  m_imageScatterPrev, m_imageScatterNext, m_imageScatterCur;
    
    string      m_strEnvMapName, m_strImageName, m_strFileExtension;
    int         m_iFrame;
    bool        m_bPlayMovie;
    vector<DisplayData> m_adataMovie;
    bool                m_bInDrawRoutine;

    void ExtractImageName( const char * in_str );
    string GetImagePath( const ImageType &in_type, const bool in_bTryAlt ) const; 
    void MakeImagePowerOfTwo( const IMAGERaw &in_image, IMAGERaw &out_image );
    void MakeImagePowerOfTwo( const IMAGERaw1D &in_image, IMAGERaw1D &out_image );
    void UndoMakeImagePowerOfTwo( const IMAGERaw &in_image, IMAGERaw &out_image, const R2Pt_i &in_iptSize );
    void UndoMakeImagePowerOfTwo( const IMAGERaw1D &in_image, IMAGERaw1D &out_image, const R2Pt_i &in_iptSize );
    void LoadScatterTexture() ;
    void LoadEnvironmentMaps() ;
    void LoadTextures() ;
    void LoadTexture1D( const ImageType &in_type ) ;
    void LoadTexture3D( const ImageType &in_type ) ;
    void SetPaper(  ) const;
    void SetLight(  ) const;
    void SetEnvMap(  ) const;
    void SetCamera(  ) const;
    void SetLightCamera( const int in_iWhich ) const;
    void DrawPolygon() const;
    void CreateHeightMap() ;
    void DrawPolygonHeightNoTex() const;
    void DrawPolygonHeight( const bool in_bDoBox ) const;
    void DrawCameraLightCircles() const;

    void DrawPaintColor( ) ;
    void DrawPaintLayer( ) ;
    void DrawHeightBW( ) ;
    void DrawHeightColor( ) ;
    void DrawImage( ) const;

    void DrawSingleLayer( ) const;
    void DrawCombinedLayers( ) const;
    
    void CalcNextRandomFrame() ;

    static void Redraw_cb( void *in_data );

    void InitializeShaders();

    double HeightScale() const ;
    int NTypes1D() const { return 4; }
    int NTypes3D() const { return 5; }
    int NImagesTotal() const { return NTypes1D() + NTypes3D(); }
    ImageType GetType1D( const int in_i ) const;
    ImageType GetType3D( const int in_i ) const;
    GLuint TextureId( const ImageType &in_type ) const;
    const R2Pt_i &GetOriginalImageSize( const ImageType &in_type ) const;
    R2Pt_i &SetOriginalImageSize( const ImageType &in_type ) ;
    IMAGERaw &Image3D( const ImageType &in_type ) ;
    const IMAGERaw &GetImage3D( const ImageType &in_type ) const;
    IMAGERaw1D &Image1D( const ImageType &in_type ) ;
    const IMAGERaw1D &GetImage1D( const ImageType &in_type ) const;
    const char *ImageName( const ImageType &in_type ) const;   

    void WriteMovieImage();

public:
    void Redraw();

    void ChangedHeight() { m_bDirtyNorm = true; }
    void ChangedLight() {  }
    void ChangedCamera() { m_bDirtyShift = true; }
    void StartRandomWalk() ;

    /**@name Read/write data */
    //@{
    ///
    void ReadShader( );
    ///
    void ReadImage( const char *in_str );
    ///
    void ReadEnvironmentMap( const char *in_str );
    ///
    void ReadImages( const char *in_str );
    ///    
    void WriteImages( const char *in_str );
    ///    
    void WriteSettings( const char *in_str );
    ///
    void MakeIrradianceMap();
    //@}

    /**@name Make movies */
    //@{
    ///
    void PlayMovie();
    ///
    void ClearMovie();
    ///
    void WriteMovie( const char *in_str );
    ///
    void ReadMovie( const char *in_str );
    //@}

    /**@name FltK overloads */
    //@{
    virtual void draw();
    virtual int handle(int);
    virtual void resize( int x, int y, int w, int h );
    //@}

    void SetUI( UserInterface *in_opUI ) { m_opUI = in_opUI; }

    Painter3D(int X, int Y, int W, int H, const char *L = NULL);
    ~Painter3D();
};
#endif
