#include <FL/gl.h>
#include "Painter3D.H"
#include "UserInterface.h"


void Painter3D::DisplayData::PullFromUI( const UserInterface *in_opUI )
{
    m_dHeightScale = in_opUI->m_dHeightScale->value();
    m_dShinyScale = in_opUI->m_dShininessScale->value();
    m_dShinyMin = in_opUI->m_dShininessStart->value();
    m_dTranspScale = in_opUI->m_dTransparencyScale->value();
    m_dDiffuseScale = in_opUI->m_dDiffusionScale->value();
    m_dRefractIndex = in_opUI->m_dSnellRefraction->value();
    m_dRefractMax = in_opUI->m_dRefractMax->value();
    m_dRefractSpeed = in_opUI->m_dRefractionFPS->value();
    m_dCameraDist = in_opUI->m_dCameraDistance->value();

    if ( in_opUI->m_bEnvironmentMap->value() ) {
        m_bUseLight = false;
    } else {
        m_bUseLight = true;
    }

    m_colLight = in_opUI->m_colLight;
    m_colShadow = in_opUI->m_colShadow;

    m_ptLight[0] = in_opUI->m_dLightLR->value();
    m_ptLight[1] = in_opUI->m_dLightUD->value();
    m_ptCamera[0] = in_opUI->m_dCameraLR->value();
    m_ptCamera[1] = in_opUI->m_dCameraUD->value();
}

void Painter3D::DisplayData::PushToUI( UserInterface *io_opUI ) const
{
    if ( !RNApproxEqual( io_opUI->m_dHeightScale->value(), m_dHeightScale ) ) {
        io_opUI->m_meshViewer->ChangedCamera();
        io_opUI->m_meshViewer->ChangedHeight();
    }
    if ( !RNApproxEqual( io_opUI->m_dCameraDistance->value(), m_dCameraDist ) ) {
        io_opUI->m_meshViewer->ChangedCamera();
    }
    if ( !RNApproxEqual( io_opUI->m_dRefractMax->value(), m_dRefractMax ) ) {
        io_opUI->m_meshViewer->ChangedCamera();
    }
    if ( !RNApproxEqual( io_opUI->m_dSnellRefraction->value(), m_dRefractIndex ) ) {
        io_opUI->m_meshViewer->ChangedCamera();
    }
    
    io_opUI->m_dHeightScale->value(m_dHeightScale);
    io_opUI->m_dTransparencyScale->value(m_dTranspScale);
    io_opUI->m_dShininessScale->value(m_dShinyScale);
    io_opUI->m_dRefractMax->value(m_dRefractMax);
    io_opUI->m_dDiffusionScale->value( m_dDiffuseScale );
    io_opUI->m_dShininessStart->value(m_dShinyMin);
    io_opUI->m_dSnellRefraction->value(m_dRefractIndex);
    io_opUI->m_dCameraDistance->value(m_dCameraDist);
    io_opUI->m_dRefractionFPS->value(m_dRefractSpeed);

    if ( m_bUseLight ) {
        io_opUI->m_bEnvironmentMap->value(0);
    } else {
        io_opUI->m_bEnvironmentMap->value(1);
    }

    io_opUI->m_colLight = m_colLight;
    io_opUI->m_colShadow = m_colShadow;
    io_opUI->ChangedColor();

    io_opUI->m_dLightLR->value(m_ptLight[0]);
    io_opUI->m_dLightUD->value(m_ptLight[1]);
    
    if ( !RNApproxEqual( io_opUI->m_dCameraLR->value(), m_ptCamera[0] ) )
        io_opUI->m_meshViewer->ChangedCamera();
    if ( !RNApproxEqual( io_opUI->m_dCameraUD->value(), m_ptCamera[1] ) )
        io_opUI->m_meshViewer->ChangedCamera();
    
    io_opUI->m_dCameraLR->value(m_ptCamera[0]);
    io_opUI->m_dCameraUD->value(m_ptCamera[1]);
}

void Painter3D::DisplayData::Write( ofstream &out ) const
{
    out << m_dHeightScale << " " << m_dTranspScale << " " << m_dDiffuseScale << " " << m_dShinyScale << "\n";
    out << m_dShinyMin << " " << m_dRefractMax << " " << m_dRefractIndex << " " << m_dRefractSpeed << "\n";
    out << m_dCameraDist << "\n";
    out << m_ptLight << m_ptCamera << "\n";
    boolWrite( out, m_bUseLight );
    m_colLight.Write(out);
    m_colShadow.Write(out);
    out << "\n";
}

void Painter3D::DisplayData::Read( ifstream &in)
{
    in >> m_dHeightScale >> m_dTranspScale >> m_dDiffuseScale >> m_dShinyScale;
    in >> m_dShinyMin >> m_dRefractMax >> m_dRefractIndex >> m_dRefractSpeed;
    in >> m_dCameraDist;
    in >> m_ptLight >> m_ptCamera;
    m_bUseLight = boolRead(in);
    m_colLight.Read(in);
    m_colShadow.Read(in);
}

bool Painter3D::DisplayData::operator==( const Painter3D::DisplayData &in_d ) const
{
    if ( m_dHeightScale != in_d.m_dHeightScale ) return false;
    if ( m_dTranspScale != in_d.m_dTranspScale ) return false;
    if ( m_dDiffuseScale != in_d.m_dDiffuseScale ) return false;
    if ( m_dShinyScale != in_d.m_dShinyScale ) return false;
    if ( m_dShinyMin != in_d.m_dShinyMin ) return false;
    if ( m_dRefractMax != in_d.m_dRefractMax ) return false;
    if ( m_dRefractIndex != in_d.m_dRefractIndex ) return false;
    if ( m_dCameraDist != in_d.m_dCameraDist ) return false;
    if ( m_ptLight[0] != in_d.m_ptLight[0] ) return false;
    if ( m_ptLight[1] != in_d.m_ptLight[1] ) return false;
    if ( m_ptCamera[0] != in_d.m_ptCamera[0] ) return false;
    if ( m_ptCamera[1] != in_d.m_ptCamera[1] ) return false;
    if ( m_bUseLight != in_d.m_bUseLight ) return false;
    if ( !(m_colLight == in_d.m_colLight) ) return false;
    if ( !(m_colShadow == in_d.m_colShadow) ) return false;
    if ( !(m_dRefractSpeed == in_d.m_dRefractSpeed) ) return false;
    return true;
}

Painter3D::DisplayData &Painter3D::DisplayData::operator=( const Painter3D::DisplayData &in_d )
{
    m_dHeightScale = in_d.m_dHeightScale;
    m_dTranspScale = in_d.m_dTranspScale;
    m_dDiffuseScale = in_d.m_dDiffuseScale;
    m_dShinyScale = in_d.m_dShinyScale;
    m_dShinyMin = in_d.m_dShinyMin;
    m_dRefractMax = in_d.m_dRefractMax;
    m_dRefractIndex = in_d.m_dRefractIndex;
    m_dCameraDist = in_d.m_dCameraDist;
    m_ptLight = in_d.m_ptLight;
    m_ptCamera = in_d.m_ptCamera;
    m_bUseLight = in_d.m_bUseLight;
    m_colLight = in_d.m_colLight;
    m_colShadow = in_d.m_colShadow;
    m_dRefractSpeed = in_d.m_dRefractSpeed;

    return *this;
}

Painter3D::DisplayData::DisplayData( const Painter3D::DisplayData &in_data )
{
    *this = in_data;
}

Painter3D::DisplayData::DisplayData()
{
    m_dHeightScale = 0.01;
    m_dTranspScale = 0.5;
    m_dDiffuseScale = 0.5;
    m_dShinyScale = 0.5;
    m_dShinyMin = 0.0;
    m_dRefractMax = 0.5;
    m_dRefractIndex = 0.5;
    m_dCameraDist = 5.0;
    m_ptLight = R2Pt(0,0);
    m_ptCamera = R2Pt(0,0);
    m_bUseLight = true;
    m_colLight = UTILSColor(1,1,1);
    m_colShadow = UTILSColor(0,0,0);
    m_dRefractSpeed = 0.1;
}

