
#include <FL/gl.h>
#include "Painter3D.H"
#include "UserInterface.h"

void Painter3D::Redraw()
{
    s_iCountRedraw = 2;

    Fl::add_timeout( 0.00001, Painter3D::Redraw_cb, this );

    redraw();

    /*
    if ( m_opUI && m_opUI->m_bRecordMovie->value() ) {
        if ( m_adataMovie.size() == 0 ) {
            m_adataMovie.add(1);
            m_adataMovie.last().PullFromUI( m_opUI );
        } else {
            DisplayData data;
            data.PullFromUI( m_opUI );
            if ( ! ( data == m_adataMovie.last() ) )
                m_adataMovie += data;
        }
    }
     */
}

void Painter3D::LoadTextures()
{
    if ( m_acTextureId.size() == 0 ) {
        m_acTextureId.resize( NImagesTotal() - 2 );
        glGenTextures( m_acTextureId.size(), &m_acTextureId[0] );
    } else {
        glDeleteTextures( m_acTextureId.size(), &m_acTextureId[0] );
        glGenTextures( m_acTextureId.size(), &m_acTextureId[0] );
    }

    for ( int iT = 0; iT < NTypes3D(); iT++ ) {
        if ( GetType3D(iT) == ENVIRONMENT_MAP )
            continue;
        if ( GetType3D(iT) == IRRADIANCE_MAP )
            continue;

        glBindTexture( GL_TEXTURE_2D, TextureId( GetType3D(iT) ) );

		//glTexParam should be placed AFTER glBindTexture
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

        IMAGERaw &image = Image3D( GetType3D(iT) );
		glTexImage2D( GL_TEXTURE_2D, 0, 3, image.Width(), image.Height(), 0, GL_RGB, GL_FLOAT, image.Data() );
    }

    for ( FORINT iT = 0; iT < NTypes1D(); iT++ ) {
        glBindTexture( GL_TEXTURE_2D, TextureId( GetType1D(iT) ) );

		//glTexParam should be placed AFTER glBindTexture
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
		glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
		glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

        IMAGERaw1D &image = Image1D( GetType1D(iT) );
		glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE8, image.Width(), image.Height(), 0, GL_LUMINANCE, GL_FLOAT, image.Data() );
    }
    LoadEnvironmentMaps();
    LoadScatterTexture();
}

void Painter3D::LoadTexture1D( const ImageType &in_type ) 
{
    glBindTexture( GL_TEXTURE_2D, TextureId( in_type ) );

	//glTexParam should be placed AFTER glBindTexture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    IMAGERaw1D &image = Image1D( in_type );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE8, image.Width(), image.Height(), 0, GL_LUMINANCE, GL_FLOAT, image.Data() );
}

void Painter3D::LoadTexture3D( const ImageType &in_type ) 
{
    glBindTexture( GL_TEXTURE_2D, TextureId( in_type ) );

	//glTexParam should be placed AFTER glBindTexture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    IMAGERaw &image = Image3D( in_type );
	glTexImage2D( GL_TEXTURE_2D, 0, 3, image.Width(), image.Height(), 0, GL_RGB, GL_FLOAT, image.Data() );
}


void Painter3D::LoadEnvironmentMaps()
{
    if ( s_iCountRedraw == 101 )
        return;
        
    if ( m_acEnvMapId.size() == 0 ) {
        m_acEnvMapId.resize( 2 );
        glGenTextures( m_acEnvMapId.size(), &m_acEnvMapId[0] );
    } else {
        glDeleteTextures( m_acEnvMapId.size(), &m_acEnvMapId[0] );
        glGenTextures( m_acEnvMapId.size(), &m_acEnvMapId[0] );
    }

    glBindTexture( GL_TEXTURE_2D, TextureId( ENVIRONMENT_MAP ) );

	//glTexParam should be placed AFTER glBindTexture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    IMAGERaw &image = Image3D( ENVIRONMENT_MAP );
	glTexImage2D( GL_TEXTURE_2D, 0, 3, image.Width(), image.Height(), 0, GL_RGB, GL_FLOAT, image.Data() );

    glBindTexture( GL_TEXTURE_2D, TextureId( IRRADIANCE_MAP ) );

	//glTexParam should be placed AFTER glBindTexture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    IMAGERaw &imageIrr = Image3D( IRRADIANCE_MAP );
	glTexImage2D( GL_TEXTURE_2D, 0, 3, imageIrr.Width(), imageIrr.Height(), 0, GL_RGB, GL_FLOAT, imageIrr.Data() );
}

void Painter3D::LoadScatterTexture() 
{
    if ( m_cScatterId != 0 ) {
        glDeleteTextures( 1, &m_cScatterId );
    }
    glGenTextures( 1, &m_cScatterId );

    glBindTexture( GL_TEXTURE_2D, m_cScatterId );

	//glTexParam should be placed AFTER glBindTexture
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST );
	glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST );
	glTexEnvi( GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL );

    if ( m_dPercScatter > m_opUI->m_dRefractionFPS->value() ) {
        m_imageScatterPrev = m_imageScatterNext;
        UTILSColor col(0,0,0);
        //TRACE("Generating new\n");
        for ( int i = 0; i < m_imageScatterNext.Width(); i++ ) {
            for ( int j = 0; j < m_imageScatterNext.Height(); j++ ) {
                col[0] = (double) ( rand() % 1024 ) / 1024.0;
                col[1] = (double) ( rand() % 1024 ) / 1024.0;
                m_imageScatterNext.Set( i, j, col );
                ASSERT( col[0] < 1.0 && col[1] < 1.0 );
            }
        }        
        m_dPercScatter = 0.0;
    }
    
    for ( int i = 0; i < m_imageScatterCur.Width(); i++ ) {
        for ( int j = 0; j < m_imageScatterCur.Height(); j++ ) {
            const double dPerc = m_dPercScatter / m_opUI->m_dRefractionFPS->value();
            m_imageScatterCur.Set( i, j, m_imageScatterPrev(i,j) * (1.0 - dPerc) + m_imageScatterNext(i,j) * ( dPerc ) );
        }
    }

	glTexImage2D( GL_TEXTURE_2D, 0, 3, m_imageScatterCur.Width(), m_imageScatterCur.Height(), 0, GL_RGB, GL_FLOAT, m_imageScatterCur.Data() );

    //TRACE("L");
    m_opShader->SetTextureBySemantic("ScatterTexture", m_cScatterId );
}


void Painter3D::DrawCameraLightCircles() const
{
    glPushAttrib( GL_LIGHTING_BIT | GL_LINE_BIT );
    glDisable( GL_LIGHTING );
    glDisable( GL_TEXTURE_2D );
    glLineWidth( 4.0 );
    OGLBegin2D( w(), h() );
    if ( m_opUI->m_bShowCamera->value() ) {
        glColor4f( 1.0, 1.0, 1.0, 1.0 );
        OGLDrawCircle( R2Pt( m_opUI->m_dCameraLR->value(), -m_opUI->m_dCameraUD->value() ), 0.02 - m_opUI->m_dCameraDistance->value() / 20.0 * 0.01 );
        glColor4f( 0.0, 0.0, 0.0, 1.0 );
        OGLDrawCircle( R2Pt( m_opUI->m_dCameraLR->value(), -m_opUI->m_dCameraUD->value() ), 0.005 );
    }
    if ( m_opUI->m_bShowLight->value() ) {
        glColor4f( m_opUI->m_colLight[0], m_opUI->m_colLight[1], m_opUI->m_colLight[2], 1.0 );
        OGLDrawCircle( R2Pt( m_opUI->m_dLightLR->value(), -m_opUI->m_dLightUD->value() ), 0.01 );
        glColor4f( m_opUI->m_colShadow[0], m_opUI->m_colShadow[1], m_opUI->m_colShadow[2], 1.0 );
        OGLDrawCircle( R2Pt( m_opUI->m_dLightLR->value(), -m_opUI->m_dLightUD->value() ), 0.005 );
    }

    OGLEnd2D();
    glPopAttrib();
}

void Painter3D::SetPaper() const
{
	m_opShader->SetTextureBySemantic("PaperTexture", TextureId( PAPER ) );
}

void Painter3D::SetLight( ) const
{
    float afPosLight[3], afColLight[3], afColShadow[3];
    const float fXLight = m_opUI->m_dLightLR->value();
    const float fYLight = -m_opUI->m_dLightUD->value();

    const double dRadius = sqrt( fXLight * fXLight + fYLight * fYLight );
    const double dX = (dRadius > 0.9) ? 0.9 * fXLight / dRadius : fXLight;
    const double dY = (dRadius > 0.9) ? 0.9 * fYLight / dRadius : fYLight;
    const double dDenom = 1.0 + dX * dX + dY * dY;
    afPosLight[0] = 10.0 * (2.0 * dX / dDenom);
    afPosLight[1] = 10.0 * (2.0 * dY / dDenom);
    afPosLight[2] = -10.0 * ( (-1.0 + dX * dX + dY * dY) / dDenom );

    afColLight[0] = m_opUI->m_colLight[0];
    afColLight[1] = m_opUI->m_colLight[1];
    afColLight[2] = m_opUI->m_colLight[2];

    afColShadow[0] = m_opUI->m_colShadow[0];
    afColShadow[1] = m_opUI->m_colShadow[1];
    afColShadow[2] = m_opUI->m_colShadow[2];

	m_opShader->SetFloat3BySemantic("LightPosition", afPosLight );
	m_opShader->SetFloat3BySemantic("LightColor", afColLight );
	m_opShader->SetFloat3BySemantic("ShadowColor", afColShadow );

    m_opShader->SetBoolBySemantic("UseDiffuse", m_opUI->m_bDiffuse->value() ? true : false );
    m_opShader->SetBoolBySemantic("UseSpecular", m_opUI->m_bSpecular->value() ? true : false );
    //m_opShader->SetBoolBySemantic("UseShadow", m_opUI->m_bDoShadow->value() ? true : false );
    m_opShader->SetBoolBySemantic("UseShadow", false );
}

void Painter3D::SetEnvMap( ) const
{
	m_opShader->SetTextureBySemantic("EnvironmentMap", TextureId( ENVIRONMENT_MAP ) );
	m_opShader->SetTextureBySemantic("IrradianceMap", TextureId( IRRADIANCE_MAP ) );
    m_opShader->SetBoolBySemantic("UseDiffuse", m_opUI->m_bDiffuse->value() ? true : false );
    m_opShader->SetBoolBySemantic("UseSpecular", m_opUI->m_bSpecular->value() ? true : false );
}

void Painter3D::SetCamera( ) const
{
    // Looking down the z axis
    float afPosCam[3];
    const float fXCam = m_opUI->m_dCameraLR->value();
    const float fYCam = -m_opUI->m_dCameraUD->value();

    const double dDist = m_opUI->m_dCameraDistance->value();
    const double dRadius = sqrt( fXCam * fXCam + fYCam * fYCam );
    const double dX = (dRadius > 0.9) ? 0.9 * fXCam / dRadius : fXCam;
    const double dY = (dRadius > 0.9) ? 0.9 * fYCam / dRadius : fYCam;
    const double dDenom = 1.0 + dX * dX + dY * dY;
    afPosCam[0] = dDist * (2.0 * dX / dDenom);
    afPosCam[1] = dDist * (2.0 * dY / dDenom);
    afPosCam[2] = -dDist * ( (-1.0 + dX * dX + dY * dY) / dDenom );

    m_opShader->SetFloat3BySemantic("CameraPosition", afPosCam );

    static R4Matrix s_matCam;
    static bool s_bCamInit = false;
    if ( s_bCamInit == false ) {
        s_bCamInit = true;
        s_matCam.MakeIdentity();
        const double dAspect = (h() == 0 ) ? 1.0 : (double) w() / (double) h();
        
        const double dWidth = 1.0 / tan( atan( 1.0 / 100.0 ) );
        const double dHeight = dAspect * dWidth;
        const double dK = -0.001 / 1000.0;
            
        s_matCam(3, 3) = 0;
        s_matCam(3, 2) = -1.0;
        s_matCam(2, 3) = dK / (1.0 + dK);
        s_matCam(2, 2) = -1.0 / (1.0 + dK);
        
        s_matCam = s_matCam *
            R4Matrix::Scaling( dWidth / 1000.0, dHeight / 1000.0, 1.0 / 1000.0, 1.0 ) *
            R4Matrix::Translation( 0, 0, -100.0 );
    }
    m_opShader->SetOGLMatrix4x4BySemantic( "CameraProjection", s_matCam );
}

void Painter3D::DrawPolygon() const
{
    OGLBegin2D( w(), h() );

    glColor4f( 0.2f, 0.2f, 0.2f, 1.0f );

    glBegin( GL_POLYGON );
    glNormal3d( 0.0, 0.0, 1.0 );

    glColor4f( 0.0, 0.0, 0,0);
    glTexCoord2d( 0.0, 0.0 );
    glVertex3d( -1.0, -1.0, 0.0 );
    glColor4f( 1.0, 0.0, 0,0);
    glTexCoord2d( 1.0, 0.0 );
    glVertex3d(  1.0, -1.0, 0.0 );
    glColor4f( 1.0, 1.0, 0,0);
    glTexCoord2d( 1.0, 1.0 );
    glVertex3d(  1.0,  1.0, 0.0 );
    glColor4f( 0.0, 1.0, 0,0);
    glTexCoord2d( 0.0, 1.0 );
    glVertex3d( -1.0,  1.0, 0.0 );
    glEnd();

    OGLEnd2D();
}

void Painter3D::CreateHeightMap() 
{
    const IMAGERaw1D &imageHeight = Image1D( HEIGHT );
    const double dHMax = HeightScale();

    const int iWidthImage = GetOriginalImageSize( HEIGHT )[0];
    const int iHeightImage = GetOriginalImageSize( HEIGHT )[1];
    if ( iHeightImage == 0 || iWidthImage == 0 )
        return;
    
    const int iMaxDim = WINmin( w(), h() );
    const double dAspectRatio = (double) iWidthImage / (double) iHeightImage;
    m_iptGridSize[0] = ( iWidthImage >= iHeightImage ) ? iMaxDim : (int) ( dAspectRatio * iMaxDim );
    m_iptGridSize[1] = ( iHeightImage >= iWidthImage ) ? iMaxDim : (int) ( (double) iMaxDim / dAspectRatio );

    const double dWidthScreen = dAspectRatio * 2.0;
    const double dHeightScreen = 2.0 / dAspectRatio;
    const double dXStart = ( m_iptGridSize[0] == iMaxDim ) ? -1.0 : -1.0 + 0.5 * (2.0 - dWidthScreen);
    const double dYStart = ( m_iptGridSize[1] == iMaxDim ) ? -1.0 : -1.0 + 0.5 * (2.0 - dHeightScreen);
    const double dXStep = (2.0 * -dXStart) / ( double ) m_iptGridSize[0];
    const double dYStep = (2.0 * -dYStart) / ( double ) m_iptGridSize[1];
    
    m_ptTexStart[0] = (m_iptGridSize[0] == iMaxDim) ? 0.0 : (1.0 - dAspectRatio) / 2.0;
    m_ptTexStart[1] = (m_iptGridSize[1] == iMaxDim) ? 0.0 : (1.0 - 1.0 / dAspectRatio) / 2.0;
    
    m_ptTexStep[0] = (1.0 - (2.0 * m_ptTexStart[0])) / ( m_iptGridSize[0] + 1.0 );
    m_ptTexStep[1] = (1.0 - (2.0 * m_ptTexStart[1])) / ( m_iptGridSize[1] + 1.0 );

    if ( m_bDirtyNorm || m_avecNorm.size() != ( m_iptGridSize[0] * m_iptGridSize[1] ) ) {
        m_avecNorm.resize( m_iptGridSize[0] * m_iptGridSize[1] );
        m_aptQuad.resize( m_iptGridSize[0] * m_iptGridSize[1] );
        m_aptShiftUV.resize( m_iptGridSize[0] * m_iptGridSize[1] );

        double dX = dXStart;
        double dXImage = m_ptTexStart[0];

        for ( int iX = 0; iX < m_iptGridSize[0]; iX++, dX += dXStep, dXImage += m_ptTexStep[0] ) {
            double dY =  dYStart;

            double dYImage = m_ptTexStart[1];
            for ( int iY = 0; iY < m_iptGridSize[1]; iY++, dY += dYStep, dYImage += m_ptTexStep[1] ) {
                ASSERT( dXImage >= 0.0 && dXImage <= 1.0 && dYImage >= 0.0 && dYImage <= 1.0 );
                ASSERT( dX >= -1.0 && dX <= 1.0 && dY >= -1.0 && dY <= 1.0 );
                m_aptQuad[ iX * m_iptGridSize[1] + iY ] = R3Pt_f( dX, dY, imageHeight( R2Pt(dXImage, 1.0 - dYImage) ) * dHMax );
            }
        }

        for ( FORINT iX = 0; iX < m_iptGridSize[0]-1; iX++ ) {
            for ( int iY = 0; iY < m_iptGridSize[1]-1; iY++ ) {
                const R3Pt_f &ptLL = m_aptQuad[ iX * m_iptGridSize[1] + iY ];
                const R3Pt_f &ptLR = m_aptQuad[ (iX+1) * m_iptGridSize[1] + iY ];
                const R3Pt_f &ptUL = m_aptQuad[ iX * m_iptGridSize[1] + iY+1 ];
                const R3Pt_f &ptUR = m_aptQuad[ (iX+1) * m_iptGridSize[1] + iY+1 ];

                // Cross product of ( dXStep, 0, dHX ) and ( 0, dYStep, dHY )
                m_avecNorm[ iX * m_iptGridSize[1] + iY ] = UnitSafe( Cross( ptLR - ptLL, ptUL - ptLL ) + Cross( ptUL - ptUR, ptLR - ptUR ) );
            }
        }
        m_bDirtyNorm = false;
        m_bDirtyShift = true;
    }

    if ( m_bDirtyShift == true ) {
        
        R3Pt_f ptCam(0,0,0);
        const double dXCam = m_opUI->m_dCameraLR->value();
        const double dYCam = -m_opUI->m_dCameraUD->value();

        const double dDist = m_opUI->m_dCameraDistance->value();
        const double dRadius = dXCam * dXCam + dYCam * dYCam;
        const double dX = (dRadius > 0.9 * 0.9) ? dXCam / dRadius : dXCam;
        const double dY = (dRadius > 0.9 * 0.9) ? dYCam / dRadius : dYCam;
        const double dDenom = 1.0 + dX * dX + dY * dY;
        ptCam[0] = (float) ( dDist * (2.0 * dX / dDenom) );
        ptCam[1] = (float) ( dDist * (2.0 * dY / dDenom) );
        ptCam[2] = (float) ( -dDist * ( (-1.0 + dX * dX + dY * dY) / dDenom ) );

        // The direct refraction vector from the height field
        const double dSnellRefract = m_opUI->m_dSnellRefraction->value();
        double dMaxShift = 0.0001;
        for ( int iX = 0; iX < m_iptGridSize[0]; iX++ ) {
            for ( int iY = 0; iY < m_iptGridSize[1]; iY++ ) {
                const int iIndex = iX * m_iptGridSize[1] + iY;

                const R3Pt_f &pt = m_aptQuad[iIndex];
                const R3Vec_f &vecNorm = m_avecNorm[iIndex];
                const R3Vec_f vecView = UnitSafe( pt - ptCam );
                R2Pt_f &ptShiftUV = m_aptShiftUV[iIndex];

	            const double dDotViewNorm = Dot( vecView * -1.0f, vecNorm );
	            const double dCosSqr = 1.0 - dSnellRefract * (1.0 - dDotViewNorm*dDotViewNorm);
                const double dCos = (dCosSqr < 0.0) ? 0.0 : sqrt( dCosSqr );

	            const R3Vec_f vecRefract = UnitSafe( vecView * (float) dSnellRefract + vecNorm * (float) (dSnellRefract * dDotViewNorm - dCos) ); 
                // Want a scale such that vecRefract[2] = pt[2]
                const double dScaleShift = fabs( vecRefract[2] ) < 1e-4f ? 0.0 : pt[2] / vecRefract[2];                
                ptShiftUV[0] = vecRefract[0] * dScaleShift;
                ptShiftUV[1] = vecRefract[1] * dScaleShift;

                const double dShift = pow( ptShiftUV[0], 2 ) + pow( ptShiftUV[1], 2 );
                if ( dShift > dMaxShift )
                    dMaxShift = dShift;
            }
        }
        dMaxShift = sqrt( dMaxShift );

        const double dRefractMax = m_opUI->m_dRefractMax->value() ;
        const double dScaleShift = dRefractMax / dMaxShift;
        
        double dXTex = m_ptTexStart[0];
        for ( FORINT iX = 0; iX < m_iptGridSize[0]; iX++, dXTex += m_ptTexStep[0] ) {

            double dYTex = m_ptTexStart[1];
            for ( int iY = 0; iY < m_iptGridSize[1]; iY++, dYTex += m_ptTexStep[1] ) {
                const int iIndex = iX * m_iptGridSize[1] + iY;
                R2Pt_f &ptShiftUV = m_aptShiftUV[iIndex];

                ptShiftUV[0] *= dScaleShift;
                ptShiftUV[1] *= dScaleShift;

                if ( dXTex < m_ptTexStart[0] + dRefractMax ) {
                    ptShiftUV[0] *= (dXTex - m_ptTexStart[0]) / dRefractMax;
                } else if ( dXTex > 1.0 - m_ptTexStart[0] - dRefractMax ) {
                    ptShiftUV[0] *= ((1.0 - m_ptTexStart[0] - dXTex) / dRefractMax);
                }
                if ( dYTex < m_ptTexStart[1] + dRefractMax ) {
                    ptShiftUV[1] *= (dYTex - m_ptTexStart[1]) / dRefractMax;
                } else if ( dYTex > 1.0 - m_ptTexStart[1] - dRefractMax ) {
                    ptShiftUV[1] *= ((1.0 - m_ptTexStart[1] - dYTex) / dRefractMax);
                }
            }
        }
        m_bDirtyShift = false;
    }
}

void Painter3D::DrawPolygonHeight( const bool in_bDoBox ) const
{
    glColor4f( 0.2f, 0.2f, 0.2f, 1.0f );
    
    if ( in_bDoBox )
        glEnable( GL_DEPTH_TEST );

    glBegin( GL_QUADS );
    double dXTex = m_ptTexStart[0];

    int iXIndex = 0;
    int iXIndexPlus1 = m_iptGridSize[1];
    for ( int iX = 0; iX < m_iptGridSize[0] - 1; iX++, dXTex += m_ptTexStep[0], iXIndex += m_iptGridSize[1], iXIndexPlus1 += m_iptGridSize[1] ) {
        double dYTex = m_ptTexStart[1];

        for ( int iY = 0; iY < m_iptGridSize[1] - 1; iY++, dYTex += m_ptTexStep[1] ) {
            const R3Pt_f &ptLL = m_aptQuad[ iXIndex + iY ];
            const R3Pt_f &ptLR = m_aptQuad[ iXIndexPlus1 + iY ];
            const R3Pt_f &ptUL = m_aptQuad[ iXIndex + iY+1 ];
            const R3Pt_f &ptUR = m_aptQuad[ iXIndexPlus1 + iY+1 ];

            const R2Pt_f &ptTexLL = m_aptShiftUV[ iXIndex + iY ];
            const R2Pt_f &ptTexLR = m_aptShiftUV[ iXIndexPlus1 + iY ];
            const R2Pt_f &ptTexUL = m_aptShiftUV[ iXIndex + iY+1 ];
            const R2Pt_f &ptTexUR = m_aptShiftUV[ iXIndexPlus1 + iY+1 ];

            glNormal3fv( &m_avecNorm[ iX * m_iptGridSize[1] + iY ][0] );

            glColor4f( dXTex, dYTex, ptTexLL[0], ptTexLL[1] );
            glVertex3fv( &ptLL[0] );

            glColor4f( dXTex + m_ptTexStep[0], dYTex, ptTexLR[0], ptTexLR[1] );
            glVertex3fv( &ptLR[0] );

            glColor4f( dXTex + m_ptTexStep[0], dYTex + m_ptTexStep[1], ptTexUR[0], ptTexUR[1] );
            glVertex3fv( &ptUR[0] );

            glColor4f( dXTex, dYTex + m_ptTexStep[1], ptTexUL[0], ptTexUL[1] );
            glVertex3fv( &ptUL[0] );
        }
    }

    if ( in_bDoBox ) {
        glColor4f( 0, 0, 0, 0 );
        
        const int iWidth = m_iptGridSize[0];
        const int iHeight = m_iptGridSize[1];
        
        // Bottom of box
        glNormal3f( 0, 0, -1 );
        glColor4f( 0, 0, 0, 0 );
        glVertex3f( m_aptQuad[ 0 ][0], m_aptQuad[ 0 ][1], -1.0 );
        glColor4f( 0, 1, 0, 0 );
        glVertex3f( m_aptQuad[ iHeight - 1 ][0], m_aptQuad[ iHeight - 1 ][1], -1.0 );
        glColor4f( 1, 1, 0, 0 );
        glVertex3f( m_aptQuad[ (iWidth-1) * iHeight + iHeight - 1 ][0], m_aptQuad[ (iWidth-1) * iHeight + iHeight - 1 ][1], -1.0 );
        glColor4f( 1, 0, 0, 0 );
        glVertex3f( m_aptQuad[ (iWidth-1) * iHeight ][0], m_aptQuad[ (iWidth-1) * iHeight ][1], -1.0 );

        // Left and right sides of box
        dXTex = 0.5 * m_ptTexStep[0];
        for ( FORINT iX = 0; iX < iWidth - 1; iX++, dXTex += m_ptTexStep[0] ) {
            glNormal3f( -1, 0, 0 );
            glColor4f( dXTex, 0, 0, 0 );
            glVertex3fv( &m_aptQuad[ iX * iHeight ][0] );
            glColor4f( dXTex, 0, 0, 0 );
            glVertex3f( m_aptQuad[ iX * iHeight ][0], m_aptQuad[ iX * iHeight ][1], -1.0 );
            glColor4f( dXTex + m_ptTexStep[0], 0, 0, 0 );
            glVertex3f( m_aptQuad[ (iX+1) * iHeight ][0], m_aptQuad[ (iX+1) * iHeight ][1], -1.0 );
            glColor4f( dXTex + m_ptTexStep[0], 0, 0, 0 );
            glVertex3fv( & m_aptQuad[ (iX+1) * iHeight ][0] );

            glNormal3f( 1, 0, 0 );
            glColor4f( dXTex, 1, 0, 0 );
            glVertex3fv( &m_aptQuad[ iX * iHeight + iHeight - 1 ][0] );
            glColor4f( dXTex + m_ptTexStep[0], 1, 0, 0 );
            glVertex3fv( & m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][0] );
            glColor4f( dXTex + m_ptTexStep[0], 1, 0, 0 );
            glVertex3f( m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][0], m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][1], -1.0 );
            glColor4f( dXTex, 1, 0, 0 );
            glVertex3f( m_aptQuad[ iX * iHeight + iHeight - 1 ][0], m_aptQuad[ iX * iHeight + iHeight - 1 ][1], -1.0 );
        }

        double dYTex = 0.5 * m_ptTexStep[1];
        for ( int iY = 0; iY < iHeight - 1; iY++, dYTex += m_ptTexStep[1] ) {
            glNormal3f( 0, -1, 0 );
            glColor4f( 0, dYTex, 0, 0 );
            glVertex3fv( &m_aptQuad[ iY ][0] );
            glColor4f( 0, dYTex, 0, 0 );
            glVertex3f( m_aptQuad[ iY ][0], m_aptQuad[ iY ][1], -1.0 );
            glColor4f( 0, dYTex + m_ptTexStep[1], 0, 0 );
            glVertex3f( m_aptQuad[ iY + 1 ][0], m_aptQuad[ iY + 1 ][1], -1.0 );
            glColor4f( 0, dYTex + m_ptTexStep[1], 0, 0 );
            glVertex3fv( & m_aptQuad[ iY + 1 ][0] );

            glNormal3f( 0, 1, 0 );
            glColor4f( 1, dYTex, 0, 0 );
            glVertex3fv( &m_aptQuad[ (iHeight-1) * iHeight + iY ][0] );
            glColor4f( 1, dYTex + m_ptTexStep[1], 0, 0 );
            glVertex3fv( & m_aptQuad[ (iHeight-1) * iHeight + iY + 1][0] );
            glColor4f( 1, dYTex + m_ptTexStep[1], 0, 0 );
            glVertex3f( m_aptQuad[  (iHeight-1) * iHeight + iY + 1 ][0], m_aptQuad[  (iHeight-1) * iHeight + iY + 1 ][1], -1.0 );
            glColor4f( 1, dYTex, 0, 0 );
            glVertex3f( m_aptQuad[  (iHeight-1) * iHeight + iY  ][0], m_aptQuad[ (iHeight-1) * iHeight + iY  ][1], -1.0 );
        }
    }
    glEnd();

    OGLEnd2D();
}

void Painter3D::DrawPolygonHeightNoTex() const
{
    glColor4f( 0.2f, 0.2f, 0.2f, 1.0f );

    const int iWidth = m_iptGridSize[0];
    const int iHeight = m_iptGridSize[1];
    
    glEnable( GL_DEPTH_TEST );

    glBegin( GL_QUADS );

    int iXIndex = 0;
    int iXIndexPlus1 = iHeight;
    for ( int iX = 0; iX < iWidth - 1; iX++, iXIndex += iHeight, iXIndexPlus1 += iHeight ) {

        for ( int iY = 0; iY < iHeight - 1; iY++ ) {
            const R3Pt_f &ptLL = m_aptQuad[ iXIndex + iY ];
            const R3Pt_f &ptLR = m_aptQuad[ iXIndexPlus1 + iY ];
            const R3Pt_f &ptUL = m_aptQuad[ iXIndex + iY+1 ];
            const R3Pt_f &ptUR = m_aptQuad[ iXIndexPlus1 + iY+1 ];

            glNormal3fv( &m_avecNorm[ iX * iHeight + iY ][0] );

            glVertex3fv( &ptLL[0] );

            glVertex3fv( &ptLR[0] );

            glVertex3fv( &ptUR[0] );

            glVertex3fv( &ptUL[0] );
        }
    }
    // Bottom of box
    glNormal3f( 0, 0, -1 );
    glVertex3f( m_aptQuad[ 0 ][0], m_aptQuad[ 0 ][1], -1.0 );
    glVertex3f( m_aptQuad[ iHeight - 1 ][0], m_aptQuad[ iHeight - 1 ][1], -1.0 );
    glVertex3f( m_aptQuad[ (iWidth-1) * iHeight + iHeight - 1 ][0], m_aptQuad[ (iWidth-1) * iHeight + iHeight - 1 ][1], -1.0 );
    glVertex3f( m_aptQuad[ (iWidth-1) * iHeight ][0], m_aptQuad[ (iWidth-1) * iHeight ][1], -1.0 );

    // Left and right sides of box

    for ( FORINT iX = 0; iX < iWidth - 1; iX++ ) {
        glNormal3f( -1, 0, 0 );
        glVertex3fv( &m_aptQuad[ iX * iHeight ][0] );
        glVertex3f( m_aptQuad[ iX * iHeight ][0], m_aptQuad[ iX * iHeight ][1], -1.0 );
        glVertex3f( m_aptQuad[ (iX+1) * iHeight ][0], m_aptQuad[ (iX+1) * iHeight ][1], -1.0 );
        glVertex3fv( & m_aptQuad[ (iX+1) * iHeight ][0] );

        glNormal3f( 1, 0, 0 );
        glVertex3fv( &m_aptQuad[ iX * iHeight + iHeight - 1 ][0] );
        glVertex3fv( & m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][0] );
        glVertex3f( m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][0], m_aptQuad[ (iX+1) * iHeight + iHeight - 1 ][1], -1.0 );
        glVertex3f( m_aptQuad[ iX * iHeight + iHeight - 1 ][0], m_aptQuad[ iX * iHeight + iHeight - 1 ][1], -1.0 );
    }


    for ( int iY = 0; iY < iHeight - 1; iY++ ) {
        glNormal3f( 0, -1, 0 );
        glVertex3fv( &m_aptQuad[ iY ][0] );
        glVertex3f( m_aptQuad[ iY ][0], m_aptQuad[ iY ][1], -1.0 );
        glVertex3f( m_aptQuad[ iY + 1 ][0], m_aptQuad[ iY + 1 ][1], -1.0 );
        glVertex3fv( & m_aptQuad[ iY + 1 ][0] );

        glNormal3f( 0, 1, 0 );
        glVertex3fv( &m_aptQuad[ (iWidth-1) * iHeight + iY ][0] );
        glVertex3fv( & m_aptQuad[ (iWidth-1) * iHeight + iY + 1][0] );
        glVertex3f( m_aptQuad[  (iWidth-1) * iHeight + iY + 1 ][0], m_aptQuad[  (iWidth-1) * iHeight + iY + 1 ][1], -1.0 );
        glVertex3f( m_aptQuad[  (iWidth-1) * iHeight + iY  ][0], m_aptQuad[ (iWidth-1) * iHeight + iY  ][1], -1.0 );
    }

    glEnd();

    OGLEnd2D();
}

void Painter3D::DrawPaintColor() 
{
    const bool bUseEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;

    if ( bUseEnvMap ) {
        m_opShader->SetTechnique("PaintEnvMap");

        SetEnvMap( );
    } else {
        m_opShader->SetTechnique( "PaintLight" );

        SetLight( );
    }

    SetCamera( );

    m_opShader->SetTextureBySemantic("DiffuseTexture", TextureId( DIFFUSE ) );
    m_opShader->SetTextureBySemantic("SpecularTexture", TextureId( SPECULAR ) );
    m_opShader->SetTextureBySemantic("ShininessTexture", TextureId( SHININESS ) );

	m_opShader->StartPass();
    DrawPolygonHeight( false );
    //DrawPolygon();
    m_opShader->EndPass();
}

void Painter3D::DrawPaintLayer() 
{
    const bool bUseEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;

    if ( bUseEnvMap ) {
        m_opShader->SetTechnique("LayerEnvMap");

        SetEnvMap( );
    } else {
        m_opShader->SetTechnique( "LayerLight" );

        SetLight( );
    }

    LoadScatterTexture();
    SetPaper();
    SetCamera( );

    m_opShader->SetTextureBySemantic("DiffuseTexture", TextureId( DIFFUSE ) );
    m_opShader->SetTextureBySemantic("SpecularTexture", TextureId( SPECULAR ) );
    m_opShader->SetTextureBySemantic("ShininessTexture", TextureId( SHININESS ) );
    m_opShader->SetTextureBySemantic("TransparencyTexture", TextureId( TRANSPARENCY ) );
    m_opShader->SetTextureBySemantic("DiffusionTexture", TextureId( DIFFUSION ) );
    m_opShader->SetTextureBySemantic("HeightTexture", TextureId( HEIGHT ) );

	m_opShader->StartPass();
    DrawPolygonHeight(false);
    //DrawPolygon();
    m_opShader->EndPass();
}

void Painter3D::DrawHeightBW() 
{
    const bool bUseEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;
    
    if ( bUseEnvMap ) {
        m_opShader->SetTechnique("HeightEnvMapBW");
        
        SetEnvMap( );
    } else {
        m_opShader->SetTechnique( "HeightLightBW" );
        
        SetLight( );
    }
    
    SetCamera( );
    
    m_opShader->SetTextureBySemantic("DiffuseTexture", TextureId( DIFFUSE ) );
    m_opShader->SetTextureBySemantic("SpecularTexture", TextureId( SPECULAR ) );
    m_opShader->SetTextureBySemantic("ShininessTexture", TextureId( SHININESS ) );
    
	m_opShader->StartPass();
    DrawPolygonHeight(false);
    //DrawPolygon();
    m_opShader->EndPass();
}

void Painter3D::DrawHeightColor() 
{
    const bool bUseEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;
    
    if ( bUseEnvMap ) {
        m_opShader->SetTechnique("HeightEnvMapColor");
        
        SetEnvMap( );
    } else {
        m_opShader->SetTechnique( "HeightLightColor" );
        
        SetLight( );
    }
    
    SetCamera( );
    
    m_opShader->SetTextureBySemantic("DiffuseTexture", TextureId( DIFFUSE ) );
    m_opShader->SetTextureBySemantic("SpecularTexture", TextureId( SPECULAR ) );
    m_opShader->SetTextureBySemantic("ShininessTexture", TextureId( SHININESS ) );
    
	m_opShader->StartPass();
    DrawPolygonHeight(false);
    //DrawPolygon();
    m_opShader->EndPass();
}

void Painter3D::DrawImage( ) const
{
    m_opShader->SetTechnique( "Image" );
    
    SetCamera( );
    m_opShader->SetTextureBySemantic("DiffuseTexture", TextureId( DIFFUSE ) );
    m_opShader->SetTextureBySemantic("SpecularTexture", TextureId( SPECULAR ) );
    m_opShader->SetTextureBySemantic("ShininessTexture", TextureId( SHININESS ) );
    m_opShader->SetTextureBySemantic("TransparencyTexture", TextureId( TRANSPARENCY ) );
    m_opShader->SetTextureBySemantic("DiffusionTexture", TextureId( DIFFUSION ) );
    m_opShader->SetTextureBySemantic("HeightTexture", TextureId( HEIGHT ) );
	m_opShader->SetTextureBySemantic("PaperTexture", TextureId( PAPER ) );
	m_opShader->SetTextureBySemantic("EnvironmentMap", TextureId( ENVIRONMENT_MAP ) );
	m_opShader->SetTextureBySemantic("IrradianceMap", TextureId( IRRADIANCE_MAP ) );
    
    m_opShader->SetIntBySemantic("SingleTexture", (int) m_opUI->m_iWhichImage->value() );

    m_opShader->StartPass();
    DrawPolygon();
    m_opShader->EndPass();
}
