// Painter3D.cpp: implementation of the Painter3D class.
//
//////////////////////////////////////////////////////////////////////

#include <FL/gl.h>
#include "Painter3D.H"
#include "UserInterface.h"

void Painter3D::ReadShader()
{
    delete m_opShader;
    m_opShader = new OGLShader2(); 
	m_opShader->LoadShaderFromFile( "../Data/Shaders/Painter3D.cgfx" );
    m_opShader->SetTechnique( (char *) m_opShader->GetTechnique(0) );
}

void Painter3D::MakeImagePowerOfTwo( const IMAGERaw &in_image, IMAGERaw &out_image )
{
    int iSize = 1;

    while ( iSize < in_image.Width() && iSize < in_image.Height() ) {
        iSize = iSize * 2;
    }

    if ( iSize == in_image.Width() && iSize == in_image.Height() ) {
        out_image = in_image;
    } else {
        out_image.SetSize( iSize, iSize );
        memset( out_image.Data(), 0, iSize * iSize * 3 * sizeof(float) );
        const int iXStart = ( iSize - in_image.Width() ) / 2;
        const int iYStart = ( iSize - in_image.Height() ) / 2;
        for ( int iX = 0; iX < in_image.Width(); iX++ ) {
            for ( int iY = 0; iY < in_image.Height(); iY++ ) {
                out_image.Set( iXStart + iX, iYStart + iY, in_image(iX, iY) );
            }
        }
    }
}

void Painter3D::UndoMakeImagePowerOfTwo( const IMAGERaw &in_image, IMAGERaw &out_image, const R2Pt_i &in_iptSize )
{
    const int iSize = in_image.Width();
    if ( iSize == in_iptSize[0] && iSize == in_iptSize[1] ) {
        out_image = in_image;
    } else {
        out_image.SetSize( in_iptSize[0], in_iptSize[1] );

        const int iXStart = ( iSize - out_image.Width() ) / 2;
        const int iYStart = ( iSize - out_image.Height() ) / 2;
        for ( int iX = 0; iX < out_image.Width(); iX++ ) {
            for ( int iY = 0; iY < out_image.Height(); iY++ ) {
                out_image.Set( iX, iY, in_image(iXStart + iX, iYStart + iY) );
            }
        }
    }
}

void Painter3D::MakeImagePowerOfTwo( const IMAGERaw1D &in_image, IMAGERaw1D &out_image )
{
    int iSize = 1;

    while ( iSize < in_image.Width() && iSize < in_image.Height() ) {
        iSize = iSize * 2;
    }

    if ( iSize == in_image.Width() && iSize == in_image.Height() ) {
        out_image = in_image;
    } else {
        out_image.SetSize( iSize, iSize );
        memset( out_image.Data(), 0, iSize * iSize * sizeof(float) );
        const int iXStart = ( iSize - in_image.Width() ) / 2;
        const int iYStart = ( iSize - in_image.Height() ) / 2;
        for ( int iX = 0; iX < in_image.Width(); iX++ ) {
            for ( int iY = 0; iY < in_image.Height(); iY++ ) {
                out_image.Set( iXStart + iX, iYStart + iY, in_image(iX, iY) );
            }
        }
    }
}

void Painter3D::UndoMakeImagePowerOfTwo( const IMAGERaw1D &in_image, IMAGERaw1D &out_image, const R2Pt_i &in_iptSize )
{
    const int iSize = in_image.Width();
    if ( iSize == in_iptSize[0] && iSize == in_iptSize[1] ) {
        out_image = in_image;
    } else {
        out_image.SetSize( in_iptSize[0], in_iptSize[1] );

        const int iXStart = ( iSize - out_image.Width() ) / 2;
        const int iYStart = ( iSize - out_image.Height() ) / 2;
        for ( int iX = 0; iX < out_image.Width(); iX++ ) {
            for ( int iY = 0; iY < out_image.Height(); iY++ ) {
                out_image.Set( iX, iY, in_image(iXStart + iX, iYStart + iY) );
            }
        }
    }
}


void Painter3D::ExtractImageName( const char * in_str )
{
    string strIn(in_str), strFileName(in_str);
    
    const size_t iExt = strIn.rfind( "." );
    if ( iExt != string::npos ) {
        m_strFileExtension = strIn.substr( iExt );
    }
    
    for ( string::iterator it=strIn.begin() ; it < strIn.end(); it++ ) {
        if ( *it == '\\' )
            *it = '/';
    }
    const size_t iSlash = strIn.rfind( "/" );
    
    if ( iSlash != string::npos ) {
        strFileName = strIn.substr( iSlash + 1 );
    }
    m_strImageName = strIn;
    
    for ( int i = DIFFUSE; i < ALL; i++ ) {
        if ( (ImageType) i == SHADOW_MAP )
            continue;

        const size_t iImageName = strFileName.rfind( ImageName( (ImageType) i ) );
        
        if ( iImageName != string::npos ) {
            if ( iSlash == string::npos )
                m_strImageName = strIn.substr( 0, iImageName - 1 );
            else
                m_strImageName = strIn.substr( 0, iSlash + iImageName );
        }
    }
    
    cout << "File name: " << m_strImageName.c_str() << "  Extension: " << m_strFileExtension.c_str() << "\n";
}

static char s_strExts[6][6] = {".tif", ".bmp", ".jpg", ".tiff", ".jpeg", ".png"};
string Painter3D::GetImagePath( const ImageType &in_type, const bool in_bTryAlt ) const
{
    string strBase = m_strImageName;
    strBase += "_";
    strBase += ImageName( in_type );
    
    string strTry = strBase;
    strTry += m_strFileExtension;
    
    ifstream inTry( strTry.c_str() );
    if ( inTry.good() ) {
        inTry.close();
        return strTry;
    }
    inTry.clear();
    
    if ( in_bTryAlt == true ) {
        for ( int i = 0; i < 6; i++ ) {
            strTry = strBase;
            strTry += s_strExts[i];
            
            inTry.open( strTry.c_str() );
            if ( inTry.good() ) {
                inTry.close();
                return strTry;
            }
            inTry.clear();
        }
    }
    
    if ( in_bTryAlt == true )
        cerr << "Unable to open file " << strBase.c_str() << " with any known image exensions\n";
    
    return strTry;
}

//////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////
/// Read a single mesh (type of read determined by file extension)
void Painter3D::ReadImages( const char *in_str )
{
    IMAGERaw image;
    IMAGERaw1D image1D;
    bool bHadEnvMap = false;

    ExtractImageName( in_str );

    //const double dScl = 796.0 / 1024.0;
    for ( int iT = 0; iT < NTypes3D(); iT++ ) {
        const string strName = GetImagePath( GetType3D(iT), true );
        if ( !image.Read( strName.c_str() ) ) {
            if ( GetType3D(iT) != ENVIRONMENT_MAP && GetType3D(iT) != IRRADIANCE_MAP )
                cerr << "Missing image " << strName.c_str() << "\n";
        } else {
            SetOriginalImageSize( GetType3D(iT) )[0] = image.Width();
            SetOriginalImageSize( GetType3D(iT) )[1] = image.Height();

            MakeImagePowerOfTwo( image, Image3D( GetType3D(iT) ) );

            if ( GetType3D(iT) == ENVIRONMENT_MAP )
                bHadEnvMap = true;
        }
    }

    for ( FORINT iT = 0; iT < NTypes1D(); iT++ ) {
        const string strName = GetImagePath( GetType1D(iT), true );
        if ( !image1D.Read( strName.c_str() ) ) {
            cerr << "Missing image " << strName.c_str() << "\n";
        } else {
            SetOriginalImageSize( GetType1D(iT) )[0] = image.Width();
            SetOriginalImageSize( GetType1D(iT) )[1] = image.Height();

            MakeImagePowerOfTwo( image1D, Image1D( GetType1D(iT) ) );
        }
    }

    const string strData = m_strImageName + ".txt";
    ifstream in( strData.c_str(), ios::in );

    if ( in.good() ) {
        DisplayData data;
        data.Read(in);
        data.PushToUI( m_opUI );

        string str;
        in >> str;
        if ( bHadEnvMap == false )
            ReadEnvironmentMap(str.c_str());
    }
    in.close();

    ChangedHeight();
    ChangedLight();

    LoadTextures();

    const int iWHMax = WINmax( w(), h() );
    const int iWidth = GetImage3D( DIFFUSE ).Width();
    const int iHeight = GetImage3D( DIFFUSE ).Height();

    if ( iWidth < iHeight ) {
        resize( x(), y(), (int) ( iWHMax * ( (double) iWidth / (double) iHeight ) ), iWHMax );
    } else if ( iHeight < iWidth ) {
        resize( x(), y(), iWHMax, (int) ( iWHMax * ( (double) iHeight / (double) iWidth ) ) );
    } else {
        resize( x(), y(), iWHMax, iWHMax );
    }

    Redraw();
}


void Painter3D::ReadEnvironmentMap( const char *in_str ) 
{
    string str( in_str );
    IMAGERaw image;
    bool bSuc = false;
    
    if ( !image.Read( in_str ) ) {
        str.insert( 0, "../Data/EnvironmentMaps/" );
        if ( !image.Read( str.c_str() ) ) {
            cerr << "Missing environment map " << in_str << "\n";
        } else {
            bSuc = true;
        }
    } else {
        bSuc = true;
    }

    if ( bSuc ) {
        m_strEnvMapName = str;

        SetOriginalImageSize( ENVIRONMENT_MAP )[0] = image.Width();
        SetOriginalImageSize( ENVIRONMENT_MAP )[1] = image.Height();
        
        MakeImagePowerOfTwo( image, Image3D( ENVIRONMENT_MAP ) );
    }
    
    const size_t iExt = str.rfind(".");
    string strIrr = str.substr( 0, iExt );
    strIrr += "_irr";
    strIrr += str.substr( iExt );

    if ( !image.Read( strIrr.c_str() ) ) {
        cerr << "Missing irradiance map " << strIrr << " : Build (y/n)?\n";
        char cAnsw;
        cin >> cAnsw;
        if ( cAnsw == 'y' ) {
            MakeIrradianceMap();
            m_imageIrrMap.Write( strIrr.c_str() );
        }
    } else {
        SetOriginalImageSize( IRRADIANCE_MAP )[0] = image.Width();
        SetOriginalImageSize( IRRADIANCE_MAP )[1] = image.Height();
        
        MakeImagePowerOfTwo( image, Image3D( IRRADIANCE_MAP ) );
    }

    LoadEnvironmentMaps();
}

void Painter3D::ReadImage( const char *in_str )
{
    const ImageType imType = (ImageType) ( (int) m_opUI->m_iWhichImage->value() );

    IMAGERaw image;
    IMAGERaw1D image1D;

    if ( imType == SHADOW_MAP )
        return;

    if ( imType == HEIGHT || imType == DIFFUSION || imType == TRANSPARENCY || imType == SHININESS ) {
        if ( !image1D.Read( in_str ) ) {
            cerr << "Couldn't open image " << in_str << "\n";
        } else {
            SetOriginalImageSize( GetType1D(imType) )[0] = image1D.Width();
            SetOriginalImageSize( GetType1D(imType) )[1] = image1D.Height();
            
            MakeImagePowerOfTwo( image1D, Image1D( GetType1D(imType) ) );
            cerr << "Replacing " << ImageName( GetType1D(imType) ) << " image with " << in_str << "\n";
        }
    } else {
        if ( !image.Read( in_str ) ) {
            cerr << "Couldn't open image " << in_str << "\n";
        } else {
            SetOriginalImageSize( GetType3D(imType) )[0] = image.Width();
            SetOriginalImageSize( GetType3D(imType) )[1] = image.Height();
            
            MakeImagePowerOfTwo( image, Image3D( GetType3D(imType) ) );
            cerr << "Replacing " << ImageName( GetType1D(imType) ) << " image with " << in_str << "\n";
        }
    }
}

/// Assumes there is a texture (.tc) file of the same name
void Painter3D::WriteImages( const char *in_str )
{
    string str;
    IMAGERaw image;
    IMAGERaw1D image2;
    
    ExtractImageName( in_str );
    for ( int iT = 0; iT < NTypes3D(); iT++ ) {
        const string strName = GetImagePath( GetType3D(iT), false );
        
        UndoMakeImagePowerOfTwo( Image3D( GetType3D(iT) ), image, GetOriginalImageSize( GetType3D(iT) ) );
        image.Write( strName.c_str() );
    }

    for ( FORINT iT = 0; iT < NTypes1D(); iT++ ) {
        const string strName = GetImagePath( GetType1D(iT), false );

        UndoMakeImagePowerOfTwo( Image1D( GetType1D(iT) ), image2, GetOriginalImageSize( GetType1D(iT) ) );
        image2.Write( strName.c_str() );
    }

    WriteSettings( in_str );
}

/// Assumes there is a texture (.tc) file of the same name
void Painter3D::WriteSettings( const char *in_str )
{
    ofstream out( in_str, ios::out );

    if ( out.good() ) {
        DisplayData data;
        data.PullFromUI( m_opUI );
        data.Write(out);

        out << m_strEnvMapName << "\n";
    }
    out.close();
}

void Painter3D::MakeIrradianceMap()
{
    const IMAGERaw &imEnv = GetImage3D( ENVIRONMENT_MAP );
    IMAGERaw &imIrr = Image3D( IRRADIANCE_MAP );
    IMAGERaw imScale( imEnv.Width(), imEnv.Height() );

    imIrr.SetSize( WINmax( imEnv.Width() / 2, 128 ), WINmax(imEnv.Height() / 2, 128) );

    UTILSColor colSum;
    R3Vec vecNorm;
    R3Vec vecDir;
    const double dXIrrStep = 2.0 / (double) imIrr.Width();
    const double dYIrrStep = 2.0 / (double) imIrr.Height();
    const double dXEnvStep = 2.0 / (double) imEnv.Width();
    const double dYEnvStep = 2.0 / (double) imEnv.Height();

    vector<R3Vec> avecSave( imEnv.Width() * imEnv.Height() );

    double dXEnv = -1.0 + dXEnvStep * 0.5;
    for ( int iX = 0; iX < imEnv.Width(); iX++, dXEnv += dXEnvStep ) {
        double dYEnv = -1.0 + dYEnvStep * 0.5;
        for ( int iY = 0; iY < imEnv.Height(); iY++, dYEnv += dYEnvStep ) {
            const double dDenom2 = 1.0 + dXEnv * dXEnv + dYEnv * dYEnv;
            if ( dDenom2 > 2.0 ) {
                vecDir = R3Vec(0,0,0);
            } else {

                // Flip vector so pointing at south pole
                vecDir[0] = -dXEnv * 2.0 / dDenom2;
                vecDir[1] = -dYEnv * 2.0 / dDenom2;
                vecDir[2] = -(-1.0 + dXEnv * dXEnv + dYEnv * dYEnv) / dDenom2;
                ASSERT( vecDir[2] > 0.0 );
                vecDir.Normalize();
            }
            avecSave[ iX * imEnv.Height() + iY ] = vecDir;
            colSum = imEnv(iX, iY);
            for ( int i = 0; i < 3; i++ ) 
                colSum[i] = pow( colSum[i], 4 );
            imScale.Set( iX, iY, colSum );
        }
    }

    double dXIrr = -1.0 + dXIrrStep * 0.5;
    for ( int iI = 0; iI < imIrr.Width(); iI++, dXIrr += dXIrrStep ) {
        double dYIrr = -1.0 + dYIrrStep * 0.5;

        for ( int iJ = 0; iJ < imIrr.Height(); iJ++, dYIrr += dYIrrStep ) {
            colSum = UTILSColor::BLACK;
            const double dDenom = 1.0 + dXIrr * dXIrr + dYIrr * dYIrr;
            if ( dDenom > 2.0 ) {
                imIrr.Set( iI, iJ, UTILSColor::BLACK );
                continue;
            }

            // Flip vector so pointing at south pole
            vecNorm[0] = -dXIrr * 2.0 / dDenom;
            vecNorm[1] = -dYIrr * 2.0 / dDenom;
            vecNorm[2] = -(-1.0 + dXIrr * dXIrr + dYIrr * dYIrr) / dDenom;
            ASSERT( vecNorm[2] > 0.0 );
            vecNorm.Normalize();

            double dSum = 0.0;
            int iOffset = 0;
            for ( int iX = 0; iX < imEnv.Width(); iX++, iOffset += imEnv.Height() ) {
                for ( int iY = 0; iY < imEnv.Height(); iY++ ) {

                    const double dPerc = Dot( vecNorm, avecSave[ iOffset + iY ] );
                    if ( dPerc > 0.0 ) {
                        colSum = colSum + imScale( iX, iY ) * dPerc;
                        dSum += 1.0;
                    }
                }
            }
            for ( int i = 0; i < 3; i++ )
                colSum[i] = pow(colSum[i] / dSum, 1.0 / 4.0);

            imIrr.Set( iI, iJ, colSum );
        }
    }
    imIrr.Write( "../Data/scenes/irr.bmp" );
}

