#include <FL/gl.h>
#include "Painter3D.H"
#include "UserInterface.h"

void Painter3D::SetLightCamera( const int in_iWhich ) const
{
    const IMAGERaw1D &im = GetImage1D( HEIGHT );
    const double dHeightMax = m_opUI->m_dHeightScale->value();
    const bool bEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;

    static OGLObjsCamera cam;
    if ( cam.Width() != im.Width() || cam.Height() != im.Height() ) {
        cam.SetSize( im.Width(), im.Height() );
        cam.SetNearFar(4.5, 15.25 );
        cam.SetZoom( atan( 1.0 / 5.0 ) * 2.0 );
    }

    R3Pt ptPos;

    if ( bEnvMap ) {
    } else {
        ASSERT( in_iWhich == 0 );

        const float fXLight = m_opUI->m_dLightLR->value();
        const float fYLight = -m_opUI->m_dLightUD->value();

        const double dRadius = sqrt( fXLight * fXLight + fYLight * fYLight );
        const double dX = (dRadius > 0.9) ? 0.9 * fXLight / dRadius : fXLight;
        const double dY = (dRadius > 0.9) ? 0.9 * fYLight / dRadius : fYLight;
        const double dDenom = 1.0 + dX *   dX + dY * dY;
        ptPos[0] = 10.0 * (2.0 * dX / dDenom);
        ptPos[1] = 10.0 * (2.0 * dY / dDenom);
        ptPos[2] = -10.0 * ( (-1.0 + dX * dX + dY * dY) / dDenom );

    }
    cam.SetFromAtUp( ptPos, R3Pt(0,0,im( im.Width() / 2, im.Height() / 2 ) * dHeightMax ), R3Vec(0,1,0) );

    m_opShader->SetOGLMatrix4x4BySemantic( "LightCamera", cam.Projection() );

    double dZMin = 1e30, dZMax = -1e30;
    double dXMin = 1e30, dXMax = -1e30;
    for ( int i = 0; i < m_aptQuad.num(); i++ ) {
        R3Pt pt( m_aptQuad[i][0], m_aptQuad[i][1], m_aptQuad[i][2] );
        const R3Pt ptProj = cam.Projection() * pt;
        dZMin = WINmin( dZMin, ptProj[2] );
        dZMax = WINmax( dZMax, ptProj[2] );
        dXMin = WINmin( dXMin, ptProj[0] );
        dXMax = WINmax( dXMax, ptProj[0] );
    }
    //TRACE("%f %f  ", dXMin, dXMax );
    //TRACE("%f %f\n", dZMin, dZMax );
}

void Painter3D::MakeShadowMaps()
{
    const bool bEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;
    if ( bEnvMap == true && m_acShadowId.num() == 4 && m_bDirtyShadow == false )
        return;

    if ( bEnvMap == false && m_acShadowId.num() == 1 && m_bDirtyShadow == false )
        return;

    //if ( !m_opUI->m_bDoShadow->value() )
        //return;

    CreateHeightMap();

    m_bDirtyShadow = false;

    m_opShader->SetTechnique( "LightDoShadowMap" );

    if ( m_acShadowId.num() != 0 ) {
        glDeleteTextures( m_acShadowId.num(), &m_acShadowId[0] );
    }
    if ( bEnvMap ) {
        m_acShadowId.need( 4 );
    } else {
        m_acShadowId.need( 1 );
    }
    glGenTextures( m_acShadowId.num(), &m_acShadowId[0] );

    const IMAGERaw1D &im = GetImage1D( HEIGHT );
    static IMAGERaw imSave( im.Width(), im.Height() );

    static const GLfloat hwBorderColor[4] = { 0.0, 0.0, 0.0, 0.0 };

    ::glViewport(0, 0, im.Width(), im.Height());
    for ( int i = 0; i < m_acShadowId.num(); i++ ) {
        glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
        glPushAttrib( GL_TEXTURE_BIT | GL_ENABLE_BIT );
        glEnable( GL_DEPTH_TEST );
        glEnable( GL_POLYGON_OFFSET_FILL );

        SetLightCamera( i );

	    m_opShader->StartPass();
        glPolygonOffset( 1.1f, 4.0f );
        DrawPolygonHeightNoTex();
        m_opShader->EndPass();

        glBindTexture( GL_TEXTURE_2D, m_acShadowId[i] );
        glTexParameterfv( GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, hwBorderColor);
        glTexParameteri( GL_TEXTURE_2D,  GL_TEXTURE_MIN_FILTER, GL_NEAREST );
        glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_RGB, 0, 0, im.Width(), im.Height(), 0);
        //OGLImages_GetImage( imSave );
        //imSave.Write( WINDataHome("3DPaintings/Video/shadow.tiff") );
        //glCopyTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, 0, 0, im.Width(), im.Height(), 0);

        glPopAttrib();
    }
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    ::glViewport(0, 0, w(), h());
}

void Painter3D::SetShadowMap( ) const
{
    const bool bEnvMap = m_opUI->m_bEnvironmentMap->value() ? true : false;

    if ( bEnvMap ) {
    } else  if ( m_acShadowId.num() ) {
        SetLightCamera(0);
        m_opShader->SetTextureBySemantic( "LightShadowMap", m_acShadowId[0] );
    }
}

