/*
 *  SystemSpecific.h
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef _SYSTEM_SPECIFIC_DEFINES_H_
#define _SYSTEM_SPECIFIC_DEFINES_H_

#include <fstream>
#include <iostream>
#include <cstring>
#include <cmath>
#include <vector>
#include <string>
using namespace std;


#ifndef ASSERT
#  include <assert.h>
#  ifndef NDEBUG
#    define ASSERT(f) assert(f)
#    define DEBUG
#  else
#    define ASSERT(f)
#  endif
#endif

// Note: If you are using a compiler that scopes ints inside of
// for loops, this should be FORINT int, otherwise, FORINT 
#ifdef WIN32
#define FORINT
#else
#define FORINT int
#endif

/** \defgroup SystemDefinesSTD Min/max.
 
 \ingroup SystemDefines
 
 I'm sick of trying to work around everyone's personal definitions
 of bool, min, and max. So I made my own :).
 bool is an enumerated type on suns/sgi, type bool on windows. 
 * @{
 */
///
template<class T> inline T WINmin(T a, T b) { return a<b?a:b; }
///
template<class T> inline T WINmax(T a, T b) { return a>b?a:b; }
///
template<class T> inline T WINminmax(T a, T tMin, T tMax) 
{ 
    return (a > tMax) ? tMax : ( (a < tMin) ? tMin : a ); 
}

#ifdef WIN32
#ifndef M_PI
const double M_PI =3.1415926535897932384626433832795028842;
#endif
#endif

/// Helper functions to read/write booleans in binary
inline void boolWrite( ostream &out, const bool &in_b )
{
    if ( in_b == false ) {
        out << " f ";
    } else {
        out << " t ";
    }
}

/// Helper functions to read/write booleans in binary
inline bool boolRead( istream &in )
{
    char c;
    in >> c;
    if ( c == 't' ) return true;
    else if ( c == 'f' ) return false;
    else ASSERT(false);
    
    return false;
}

/** @} */

inline void WINWriteBinary(const char *name, ofstream &out)
{ 
    out.clear();
    
#ifdef sgi
    out.open(name, ios::out); 
#else
    out.open(name, ios::out | ios::binary); 
#endif
    if (!out.good()) {
        cerr << "Bad file " << name << "\n";
        ASSERT(false); 
    }
}

///
inline void WINReadBinary(const char *name, ifstream &in)
{ 
    in.clear();
#ifdef sgi
    in.open(name, ios::in); 
#else
    in.open(name, ios::in | ios::binary); 
#endif
    
    if (!in.good()) {
        cerr << "Bad file " << name << "\n";
        ASSERT(false); 
    }
}


#endif