/*
 *  UTILSColor.h
 *  Painter3D
 *
 *  Created by Cindy Grimm on 2/27/09.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef _UTILS_COLOR_H
#define _UTILS_COLOR_H

#include "SystemSpecific.h"

/** \class UTILSColor Utils_Color.H utils/Utils_Color.H
 \ingroup UtilitiesColor
 \brief An RGB tripple.
 
 There are several default colors, or you can create your own with
 (0-1, 0-1, 0-1) RGB.
 
 <br>   Files: 
 - UtilsColor.h
 - UTILSColor.cpp
 */

class UTILSColor {
public:
    /**@name Default colors */
    //@{
    typedef enum { 
        ///
        RED=0, 
        ///
        GREEN=1, 
        ///
        BLUE=2, 
        ///
        PURPLE=3, 
        ///
        YELLOW=4, 
        ///
        WHITE=5, 
        ///
        GREY=6,
        ///
    BLACK=7} Color;
    //@}
    
protected:
    static const double sm_aadColors[ 8 ][ 3 ];
    double m_adC[ 3 ];
    
public:
    
    /**@name Accessors */
    //@{
    ///
    double           operator[  ]( int in_iI ) const ;
    
    ///
    double &         operator[  ]( int in_iI ) ;
    //@}
    
    ///
    bool         operator==( const UTILSColor & ) const;
    ///
    bool         operator==( Color ) const;
        
    /**@name Constructors */
    //@{
    ///
    UTILSColor( );
    /// RGB between 0 and 1
    UTILSColor( double, double, double );
    /// Construct with defined color
    inline UTILSColor( Color );
    ///
    UTILSColor( const UTILSColor & );
    ///
    UTILSColor operator+( const UTILSColor & ) const;
    ///
    UTILSColor operator-( const UTILSColor & ) const;
    ///
    UTILSColor operator*( double ) const;
     ///
     inline UTILSColor & operator=( Color );

    ///
    UTILSColor & operator=( const UTILSColor & );
    ///
    ~UTILSColor( ) { }
    //@}
    
    /**@name Read/write */
    //@{
    ///
    void Write( ofstream & ) const;
    ///
    bool Read( ifstream & );
};

///
ostream & operator<<( ostream & os, const UTILSColor & c );

//------------------------------------------------------------------------------------------
inline double UTILSColor::operator[  ]( int in_iI ) const 
{ 
    ASSERT( in_iI >=0 && in_iI < 3 ); 
    return m_adC[ in_iI ]; 
}

//------------------------------------------------------------------------------------------
inline double & UTILSColor::operator[  ]( int in_iI ) 
{ 
    ASSERT( in_iI>=0 && in_iI < 3 ); 
    return m_adC[ in_iI ]; 
}

//------------------------------------------------------------------------------------------
inline
UTILSColor UTILSColor::operator+( const UTILSColor &in_col ) const
{
    return UTILSColor(in_col[0] + m_adC[0], in_col[1] + m_adC[1], in_col[2] + m_adC[2] );
}

//------------------------------------------------------------------------------------------
inline
UTILSColor UTILSColor::operator-( const UTILSColor &in_col ) const
{
    return UTILSColor(m_adC[0] - in_col[0], m_adC[1] - in_col[1], m_adC[2] - in_col[2] );
}

//------------------------------------------------------------------------------------------
inline 
UTILSColor UTILSColor::operator*( double in_dScl ) const
{
    return UTILSColor( m_adC[0] * in_dScl, m_adC[1] * in_dScl, m_adC[2] * in_dScl );
}

//------------------------------------------------------------------------------------------
inline
UTILSColor & UTILSColor::operator=( UTILSColor::Color _c )
{ 
    for ( int iI = 0; iI < 3; iI++ )
        m_adC[ iI ] = sm_aadColors[ _c ][ iI ];
    
    return * this;
    
}//UTILSColor::operator=

//------------------------------------------------------------------------------------------
inline
UTILSColor::UTILSColor( UTILSColor::Color _c ) 
{
    for ( int iI = 0; iI < 3; iI++ )
        m_adC[ iI ] = sm_aadColors[ _c ][ iI ];
    
}

#endif

